#
# Copyright 2006-2009 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
XLIFF classes specifically suited for handling the PO representation in
XLIFF.

This way the API supports plurals as if it was a PO file, for example.
"""

import contextlib
from collections.abc import Iterable, Iterator
from typing import TypeVar

from lxml import etree

from translate.misc.multistring import multistring
from translate.misc.xml_helpers import parse_xml, setXMLspace
from translate.storage import base, lisa, poheader, xliff
from translate.storage.placeables import general


def hasplurals(thing):
    if not isinstance(thing, multistring):
        return False
    return len(thing.strings) > 1


class PoXliffUnit(xliff.xliffunit):
    """A class to specifically handle the plural units created from a po file."""

    rich_parsers = general.parsers
    XML_LANG = "{http://www.w3.org/XML/1998/namespace}lang"

    def __init__(self, source=None, empty=False, **kwargs) -> None:
        self._rich_source = None
        self._rich_target = None
        self._state_n = 0
        self.units = []

        if empty:
            return

        if not hasplurals(source):
            super().__init__(source)
            return

        self.xmlelement = etree.Element(self.namespaced("group"))
        self.xmlelement.set("restype", "x-gettext-plurals")
        self.source = source

    def __eq__(self, other):
        if isinstance(other, PoXliffUnit):
            if len(self.units) != len(other.units):
                return False
            if not super().__eq__(other):
                return False
            for i in range(len(self.units) - 1):
                if not self.units[i + 1] == other.units[i + 1]:
                    return False
            return True
        if len(self.units) <= 1:
            if isinstance(other, lisa.LISAunit):
                return super().__eq__(other)
            return self.source == other.source and self.target == other.target
        return False

    # XXX: We don't return language nodes correctly at the moment
    #    def getlanguageNodes(self):
    #        if not self.hasplural():
    #            return super().getlanguageNodes()
    #        else:
    #            return self.units[0].getlanguageNodes()

    @property
    def source(self):
        if not self.hasplural():
            return super().source
        return multistring([unit.source for unit in self.units])

    @source.setter
    def source(self, source) -> None:
        self.setsource(source, sourcelang="en")

    def setsource(self, source, sourcelang="en") -> None:  # ty:ignore[invalid-method-override]
        self._rich_source = None
        if not hasplurals(source):
            if self.hasplural():
                # Don't add <source> directly on the <group> element.
                # Set source on each child trans-unit instead.
                for unit in self.units:
                    unit.source = source
                # Also remove any stale <source> element directly on the group
                for child in list(self.xmlelement):
                    if child.tag == self.namespaced("source"):
                        self.xmlelement.remove(child)
                return
            super().setsource(source, sourcelang)
        else:
            target = self.target
            if self.hasplural():
                had_target = any(
                    unit.xmlelement.find(unit.namespaced("target")) is not None
                    or bool(unit.target)
                    for unit in self.units
                )
            else:
                had_target = self.xmlelement.find(
                    self.namespaced("target")
                ) is not None or bool(target)
            if not self.hasplural():
                old_element = self.xmlelement
                group = etree.Element(self.namespaced("group"))
                for key, value in old_element.attrib.items():
                    group.set(key, value)
                group.set("restype", "x-gettext-plurals")
                for child in list(old_element):
                    if child.tag not in {
                        self.namespaced("source"),
                        self.namespaced("target"),
                    }:
                        old_element.remove(child)
                        group.append(child)
                parent = old_element.getparent()
                if parent is not None:
                    parent.replace(old_element, group)
                self.xmlelement = group
            # Remove any stale <source> element directly on the group
            for child in list(self.xmlelement):
                if child.tag == self.namespaced("source"):
                    self.xmlelement.remove(child)
            for unit in self.units:
                with contextlib.suppress(ValueError):
                    self.xmlelement.remove(unit.xmlelement)
            self.units = []
            for s in source.strings:
                newunit = xliff.xliffunit(s)
                self.units.append(newunit)
                self.xmlelement.append(newunit.xmlelement)
            # Propagate group ID to new child trans-units
            group_id = self.xmlelement.get("id")
            if group_id and self.units:
                for i, unit in enumerate(self.units):
                    unit.setid(f"{group_id}[{i}]")
            if had_target and target is not None:
                self.target = target

    # We don't support any rich strings yet
    multistring_to_rich = base.TranslationUnit.multistring_to_rich  # ty:ignore[invalid-method-override]
    rich_to_multistring = base.TranslationUnit.rich_to_multistring

    rich_source = base.TranslationUnit.rich_source
    rich_target = base.TranslationUnit.rich_target

    def gettarget(self, lang=None):
        if self.hasplural():
            strings = [unit.target or "" for unit in self.units]
            if strings:
                return multistring(strings)
            return None
        return super().gettarget(lang)

    def settarget(self, target, lang="xx", append=False) -> None:
        self._rich_target = None
        if self.target == target:
            return
        if not self.hasplural():
            super().settarget(target, lang, append)
            return
        if target is None:
            # Clear targets on all child units
            for unit in self.units:
                unit.target = None
            return
        if not isinstance(target, multistring):
            target = multistring(target)
        source = self.source
        sourcel = len(source.strings)
        targetl = len(target.strings)
        if sourcel < targetl:
            sources = source.strings + [source.strings[-1]] * (targetl - sourcel)
            targets = target.strings
            id = self.getid()
            self.source = multistring(sources)
            self.setid(id)
        elif targetl < sourcel:
            targets = target.strings + [""] * (sourcel - targetl)
        else:
            targets = target.strings

        for i, unit in enumerate(self.units):
            unit.target = targets[i]

    def getalttrans(self, origin=None):
        if not self.hasplural():
            return super().getalttrans(origin=origin)

        child_alternatives = [unit.getalttrans(origin=origin) for unit in self.units]
        if not child_alternatives:
            return []

        alternatives = []
        used_alternatives = [set() for _ in child_alternatives]
        for key in self._aggregate_alttrans_keys(child_alternatives):
            plural_alternatives = []
            for child_index, child_units in enumerate(child_alternatives):
                for unit_index, unit in enumerate(child_units):
                    if self._alttrans_key(unit) == key:
                        used_alternatives[child_index].add(unit_index)
                        plural_alternatives.append(unit)
                        break
            contexts = {unit.getcontext() for unit in plural_alternatives}
            alternative = base.TranslationUnit(
                multistring([unit.source for unit in plural_alternatives])
            )
            alternative.target = multistring(
                [unit.target for unit in plural_alternatives]
            )
            if contexts:
                alternative.setcontext(contexts.pop())
            alternative._xliff_alttrans_units = plural_alternatives  # ty:ignore[unresolved-attribute]
            alternatives.append(alternative)
        for child_index, child_units in enumerate(child_alternatives):
            alternatives.extend(
                unit
                for unit_index, unit in enumerate(child_units)
                if unit_index not in used_alternatives[child_index]
            )
        return alternatives

    def _aggregate_alttrans_keys(self, child_alternatives):
        """Return unambiguous alt-trans keys present in every plural child."""
        if not child_alternatives:
            return []

        ordered_keys = []
        for alternative in child_alternatives[0]:
            key = self._alttrans_key(alternative)
            if key not in ordered_keys:
                ordered_keys.append(key)

        return [
            key
            for key in ordered_keys
            if all(
                sum(
                    1
                    for alternative in alternatives
                    if self._alttrans_key(alternative) == key
                )
                == 1
                for alternatives in child_alternatives
            )
        ]

    def _alttrans_key(self, alternative):
        node = alternative.xmlelement
        return (
            node.get("origin"),
            node.get("match-quality"),
            node.get(self.XML_LANG),
            alternative.getcontext(),
        )

    def delalttrans(self, alternative) -> None:
        """Remove an alternate translation, including aggregated plural handles."""
        for alt in getattr(alternative, "_xliff_alttrans_units", [alternative]):
            xmlelement = alt.xmlelement
            parent = xmlelement.getparent()
            if parent is not None:
                parent.remove(xmlelement)

    def addnote(self, text, origin=None, position="append") -> None:
        """Add a note specifically in a "note" tag."""
        note = etree.SubElement(self.xmlelement, self.namespaced("note"))
        note.text = text
        if origin:
            note.set("from", origin)
        for unit in self.units[1:]:
            unit.addnote(text, origin)

    def getnotes(self, origin=None):
        # NOTE: We support both <context> and <note> tags in xliff files for comments
        if origin == "translator":
            notes = super().getnotes("translator")
            trancomments = self.gettranslatorcomments()
            if notes == trancomments or trancomments.find(notes) >= 0:
                notes = ""
            elif notes.find(trancomments) >= 0:
                trancomments = notes
                notes = ""
            return trancomments + notes
        if origin in {"programmer", "developer", "source code"}:
            devcomments = super().getnotes("developer")
            autocomments = self.getautomaticcomments()
            if (
                # pylint: disable-next=chained-comparison
                devcomments != autocomments
                and autocomments.find(devcomments) < 0
                and devcomments.find(autocomments) >= 0
            ):
                autocomments = devcomments
            return autocomments
        return super().getnotes(origin)

    def markfuzzy(self, value=True) -> None:
        super().markfuzzy(value)
        for unit in self.units[1:]:
            unit.markfuzzy(value)

    def marktranslated(self) -> None:
        super().marktranslated()
        for unit in self.units[1:]:
            unit.marktranslated()

    def setid(self, id) -> None:
        super().setid(id)
        if len(self.units) > 1:
            for i, unit in enumerate(self.units):
                unit.setid(f"{id}[{i}]")

    def getlocations(self):
        """Returns all the references (source locations)."""
        groups = self.getcontextgroups("po-reference")
        references = []
        for group in groups:
            sourcefile = ""
            linenumber = ""
            for type, text in group:
                if type == "sourcefile":
                    sourcefile = text
                elif type == "linenumber":
                    linenumber = text
            assert sourcefile
            if linenumber:
                sourcefile = f"{sourcefile}:{linenumber}"
            references.append(sourcefile)
        return references

    def getautomaticcomments(self):
        """
        Returns the automatic comments (x-po-autocomment), which corresponds
        to the #. style po comments.
        """
        groups = self.getcontextgroups("po-entry")
        comments = [
            text
            for group in groups
            for ctype, text in group
            if ctype == "x-po-autocomment"
        ]
        return "\n".join(comments)

    def gettranslatorcomments(self):
        """
        Returns the translator comments (x-po-trancomment), which
        corresponds to the # style po comments.
        """
        groups = self.getcontextgroups("po-entry")
        comments = [
            text
            for group in groups
            for ctype, text in group
            if ctype == "x-po-trancomment"
        ]
        return "\n".join(comments)

    def isheader(self):
        return "gettext-domain-header" in (self.getrestype() or "")

    def istranslatable(self):
        return super().istranslatable() and not self.isheader()

    @classmethod
    def createfromxmlElement(cls, element, namespace=None):
        namespace = namespace or ""
        if element.tag.endswith("trans-unit"):
            object = cls(None, empty=True)
            object.xmlelement = element
            object.namespace = namespace
            return object
        assert element.tag.endswith("group")
        group = cls(None, empty=True)
        group.xmlelement = element
        group.namespace = namespace
        units = list(element.iterdescendants(group.namespaced("trans-unit")))
        for unit in units:
            subunit = xliff.xliffunit.createfromxmlElement(unit)
            subunit.namespace = namespace
            group.units.append(subunit)
        return group

    def hasplural(self):
        return self.xmlelement.tag == self.namespaced("group")


U = TypeVar("U", bound=PoXliffUnit)


class PoXliffFile(xliff.xlifffile[U], poheader.poheader):
    """a file for the po variant of Xliff files."""

    UnitClass = PoXliffUnit

    def __init__(self, *args, **kwargs) -> None:
        if "sourcelanguage" not in kwargs:
            kwargs["sourcelanguage"] = "en-US"
        xliff.xlifffile.__init__(self, *args, **kwargs)

    def createfilenode(
        self, filename, sourcelanguage="en-US", datatype="po"
    ) -> etree.Element:  # ty:ignore[invalid-method-override]
        # Let's ignore the sourcelanguage parameter opting for the internal
        # one. PO files will probably be one language
        return super().createfilenode(
            filename, sourcelanguage=self.sourcelanguage, datatype="po"
        )

    def _insert_header(self, header) -> None:
        header.xmlelement.set("restype", "x-gettext-domain-header")
        header.xmlelement.set("approved", "no")
        setXMLspace(header.xmlelement, "preserve")
        self.addunit(header)

    def addheaderunit(self, target, filename):
        unit = self.addsourceunit(target, filename, True)
        unit.target = target
        unit.xmlelement.set("restype", "x-gettext-domain-header")
        unit.xmlelement.set("approved", "no")
        setXMLspace(unit.xmlelement, "preserve")
        return unit

    def parse(self, xml) -> None:
        """Populates this object from the given xml string."""
        # TODO: Make more robust

        def ispluralgroup(node: etree._Element) -> bool:
            """Determines whether the xml node refers to a gettext plural."""
            return node.get("restype") == "x-gettext-plurals"

        def isunit_parse_anchor(node: etree._Element) -> bool:
            """
            Determines whether the xml node should anchor unit insertion during
            iteration.

            This includes true standalone units and the first trans-unit in each
            plural group, which acts as the placeholder where we inject the
            whole plural group back into the unit stream.
            """
            parent = node.getparent()
            if parent is None or not ispluralgroup(parent):
                return True
            first_plural_child = next(
                parent.iterchildren(self.namespaced(self.UnitClass.rootNode)), None
            )
            return node is first_plural_child

        def pluralunits(
            pluralgroups: Iterable[etree._Element],
        ) -> Iterator[PoXliffUnit]:
            for pluralgroup in pluralgroups:
                yield self.UnitClass.createfromxmlElement(
                    pluralgroup, namespace=self.namespace
                )

        def next_nonempty_plural(
            pluralunit_iter: Iterator[PoXliffUnit],
        ) -> PoXliffUnit | None:
            nextplural = next(pluralunit_iter, None)
            while nextplural is not None and not nextplural.units:
                nextplural = next(pluralunit_iter, None)
            return nextplural

        self.filename = getattr(xml, "name", "")
        if hasattr(xml, "read"):
            xml.seek(0)
            xmlsrc = xml.read()
            xml = xmlsrc
        self.document = parse_xml(xml).getroottree()
        self.initbody()
        root_node = self.document.getroot()
        assert root_node.tag == self.namespaced(self.rootNode)
        groups = root_node.iterdescendants(self.namespaced("group"))
        pluralgroups = filter(ispluralgroup, groups)
        termEntries = root_node.iterdescendants(
            self.namespaced(self.UnitClass.rootNode)
        )

        singularunits = list(filter(isunit_parse_anchor, termEntries))
        if len(singularunits) == 0:
            return
        pluralunit_iter = pluralunits(pluralgroups)
        nextplural = next_nonempty_plural(pluralunit_iter)

        for entry in singularunits:
            if (
                nextplural
                and nextplural.units
                and entry is nextplural.units[0].xmlelement
            ):
                self.addunit(nextplural, new=False)
                nextplural = next_nonempty_plural(pluralunit_iter)
            else:
                term = self.UnitClass.createfromxmlElement(
                    entry, namespace=self.namespace
                )
                self.addunit(term, new=False)



#
# Copyright 2007,2009-2011 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

r"""
Class that manages JSON data files for translation.

JSON is an acronym for JavaScript Object Notation, it is an open standard
designed for human-readable data interchange.

JSON basic types:

- Number (integer or real)
- String (double-quoted Unicode with backslash escaping)
- Boolean (true or false)
- Array (an ordered sequence of values, comma-separated and enclosed in square
  brackets)
- Object (a collection of key:value pairs, comma-separated and enclosed in
  curly braces)
- null

Example:
--------
.. code-block:: json

   {
        "firstName": "John",
        "lastName": "Smith",
        "age": 25,
        "address": {
            "streetAddress": "21 2nd Street",
            "city": "New York",
            "state": "NY",
            "postalCode": "10021"
        },
        "phoneNumber": [
            {
              "type": "home",
              "number": "212 555-1234"
            },
            {
              "type": "fax",
              "number": "646 555-4567"
            }
        ]
   }


Todo:
-----
- Handle ``\u`` and other escapes in Unicode
- Manage data type storage and conversion. True --> "True" --> True


"""

from __future__ import annotations

import json
import re
import uuid
from collections import defaultdict
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Generic,
    Never,
    NotRequired,
    TextIO,
    TypedDict,
    TypeVar,
    cast,
)

from translate.lang.data import cldr_plural_categories
from translate.misc.multistring import multistring
from translate.storage import base

if TYPE_CHECKING:
    from collections.abc import Generator


class BaseJsonUnit(base.DictUnit):
    """A JSON entry."""

    ID_FORMAT = ".{}"

    def __init__(
        self, source=None, item=None, notes=None, placeholders=None, **kwargs
    ) -> None:
        identifier = hex(hash(source)) if source else str(uuid.uuid4())
        # Global identifier across file
        self._id = self.ID_FORMAT.format(identifier)
        # Identifier at this level
        self._item = identifier if item is None else item
        # Type conversion for the unit
        self._type = str if source is None else type(source)
        if notes:
            self.notes = notes
        self.placeholders = placeholders
        if source:
            if issubclass(self._type, str):
                self.target = source
            else:
                self.target = str(source)
        super().__init__(source)

    @property
    def source(self):
        return self.target

    @source.setter
    def source(self, source) -> None:
        self.target = source

    def setid(self, value, unitid=None) -> None:
        super().setid(value, unitid)
        self.get_unitid()
        self._item = self._unitid.parts[-1][1]  # ty:ignore[unresolved-attribute]

    def getid(self):
        return self._id

    def getlocations(self):
        return [self.getid()]

    def __str__(self) -> str:
        """Converts to a string representation."""
        return json.dumps(
            self.getvalue(), separators=(",", ": "), indent=4, ensure_ascii=False
        )

    def converttarget(self):
        if self.target is None:
            return None
        try:
            return self._type(self.target)
        except ValueError:
            return str(self.target)

    def storevalues(self, output) -> None:
        self.storevalue(output, self.converttarget())


class FlatUnitId(base.UnitId):
    @classmethod
    def from_string(cls, text):
        key = text.removeprefix(".")
        return cls([("key", key)])


class FlatJsonUnit(BaseJsonUnit):
    IdClass = FlatUnitId


class DumpArgsType(TypedDict):
    separators: tuple[str, str]
    indent: int
    ensure_ascii: bool
    sort_keys: NotRequired[bool]


JsonUnit = TypeVar("JsonUnit", bound=BaseJsonUnit)


def require_json_dict(data: Any, format_name: str) -> dict[str, Any]:
    if not isinstance(data, dict):
        raise base.ParseError(f"File is not a valid {format_name} JSON file!")
    return data


def require_json_list(data: Any, format_name: str) -> list[Any]:
    if not isinstance(data, list):
        raise base.ParseError(f"File is not a valid {format_name} JSON file!")
    return data


def require_json_string(value: Any, field: str) -> str:
    if not isinstance(value, str):
        raise base.ParseError(f"{field} does not contain string: {value!r}")
    return value


class JsonFile(base.DictStore[JsonUnit], Generic[JsonUnit]):
    """A JSON file."""

    UnitClass: type[JsonUnit] = FlatJsonUnit  # ty:ignore[invalid-assignment]

    def __init__(self, inputfile=None, filter=None, **kwargs) -> None:
        """Construct a JSON file, optionally reading in from inputfile."""
        super().__init__(**kwargs)
        self._filter = filter
        self.filename = ""
        self._file = ""
        self.dump_args: DumpArgsType = {
            "separators": (",", ": "),
            "indent": 4,
            "ensure_ascii": False,
        }
        if inputfile is not None:
            self.parse(inputfile)

    def serialize(self, out) -> None:
        units = self.get_root_node()
        self.serialize_units(units)
        out.write(json.dumps(units, **self.dump_args).encode(self.encoding))
        out.write(b"\n")

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        """
        Recursive function to extract items from the data files.

        :param data: the current branch to walk down
        :param stop: a list of leaves to extract or None to extract everything
        :param prev: the hierarchy of the tree at this iteration
        :param name_node:
        :param name_last_node: the name of the last node
        :param last_node: the last list or dict
        """
        if prev is None:
            prev = self.UnitClass.IdClass([])
        if isinstance(data, dict):
            for k, v in data.items():
                yield from self._extract_units(
                    v, stop, prev.extend("key", k), k, None, data
                )
        elif isinstance(data, list):
            for i, item in enumerate(data):
                yield from self._extract_units(
                    item, stop, prev.extend("index", i), i, name_node, data
                )
        # apply filter
        elif prev.parts and (
            stop is None
            or (isinstance(last_node, dict) and name_node in stop)
            or (isinstance(last_node, list) and name_last_node in stop)
        ):
            unit = self.UnitClass(data, name_node)
            unit.set_unitid(prev)
            yield unit

    def preprocess_input(self, text: str) -> str:
        return text

    def parse(self, data: str | bytes | TextIO | BinaryIO) -> None:
        """Parse the given file or file source string."""
        text: str | bytes
        if hasattr(data, "name"):
            self.filename = data.name
        elif not getattr(self, "filename", ""):
            self.filename = ""
        if hasattr(data, "read"):
            # Make type checking happy
            data = cast("BinaryIO", data)

            text = data.read()
            data.close()
        else:
            text = data

        if isinstance(text, bytes):
            # The JSON files should be UTF-8, but implementations
            # that parse JSON texts MAY ignore the presence of a byte order mark
            # rather than treating it as an error, see RFC7159
            decoded, self.encoding = self.detect_encoding(text)
            if decoded is None:
                raise base.ParseError(ValueError("Failed to decode JSON string."))
            text = decoded
        text = self.preprocess_input(text)
        try:
            self._file = json.loads(text)
        except ValueError as e:
            raise base.ParseError(e) from e

        for unit in self._extract_units(self._file, stop=self._filter):
            self.addunit(unit)


class JsonNestedUnit(BaseJsonUnit):
    """A nested JSON entry."""


class JsonNestedFile(JsonFile):
    """A JSON file with nested keys."""

    UnitClass = JsonNestedUnit


class WebExtensionJsonUnit(FlatJsonUnit):
    def storevalues(self, output) -> None:
        value = {"message": self.target}
        if self.notes:
            value["description"] = self.notes
        if self.placeholders:
            value["placeholders"] = self.placeholders
        self.storevalue(output, value)


class WebExtensionJsonFile(JsonFile):
    """
    WebExtension JSON file.

    See following URLs for doc:

    https://developer.chrome.com/extensions/i18n
    https://developer.mozilla.org/en-US/Add-ons/WebExtensions/Internationalization
    """

    UnitClass = WebExtensionJsonUnit

    COMMENT_RE = re.compile(r'^((?:[^"\n]|"(?:[^"\\\n]|\\.)*")*?)//.*', re.MULTILINE)

    def preprocess_input(self, text: str) -> str:
        return self.COMMENT_RE.sub(r"\1", text)

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        for item, value in require_json_dict(data, "WebExtension").items():
            if not isinstance(value, dict):
                raise base.ParseError("File is not a valid WebExtension JSON file!")
            message = value.get("message", "")
            description = value.get("description", "")
            placeholders = value.get("placeholders", None)
            message = require_json_string(message, f"Key {item!r}")
            description = require_json_string(description, f"Key {item!r} description")
            if placeholders is not None and not isinstance(placeholders, dict):
                raise base.ParseError(
                    f"Key {item!r} placeholders do not contain object: {placeholders!r}"
                )
            unit = self.UnitClass(
                message,
                item,
                description,
                placeholders,
            )
            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
            yield unit


class I18NextUnit(JsonNestedUnit):
    """
    A i18next v3 format, JSON with plurals.

    See https://www.i18next.com/
    """

    @staticmethod
    def _is_valid_suffix(suffix: str) -> bool:
        return suffix == "0"

    def _get_base_name(self):
        """Return base name for plurals."""
        item = self._item[0]
        if "_" in item:
            plural_base, _sep, suffix = item.rpartition("_")
            if self._is_valid_suffix(suffix):
                return plural_base
        return item

    def _get_plural_labels(self, count):
        base_name = self._get_base_name()
        if count == 2:
            return [base_name, f"{base_name}_plural"][:count]
        return [f"{base_name}_{i}" for i in range(count)]

    def _fixup_item(self) -> None:
        if isinstance(self._target, multistring):
            count = len(self._target.strings)
            is_list = isinstance(self._item, list)
            if not is_list or count != len(self._item):
                if not is_list:
                    self._item = [self._item]
                # Generate new plural labels
                self._item = self._get_plural_labels(count)
        elif isinstance(self._item, list):
            # Changing plural to singular
            self._item = self._get_base_name()

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, target) -> None:
        self._rich_target = None
        self._target = target

    def storevalues(self, output) -> None:
        if not isinstance(self.target, multistring):
            super().storevalues(output)
        else:
            plural_tags = self._store.get_plural_tags(self.target)  # ty:ignore[unresolved-attribute]
            if len(self.target.strings) > len(plural_tags):
                self.target.extra_strings = self.target.extra_strings[
                    : len(plural_tags) - 1
                ]
            self._fixup_item()
            for i, value in enumerate(self.target.strings):
                self.storevalue(output, value, override_key=self._item[i])


class I18NextFile(JsonNestedFile):
    """
    A i18next v3 format, this is nested JSON with several additions.

    See https://www.i18next.com/
    """

    UnitClass = I18NextUnit

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        if prev is None:
            prev = self.UnitClass.IdClass([])
        if isinstance(data, dict):
            plurals_multiple = [
                key.rsplit("_", 1)[0] for key in data if key.endswith("_0")
            ]
            plurals_simple = [
                key.rsplit("_", 1)[0] for key in data if key.endswith("_plural")
            ]
            processed = set()

            for k, v in data.items():
                # Check already processed items
                if k in processed:
                    continue
                plurals = []
                plural_base = ""
                if k in plurals_simple or f"{k}_plural" in plurals_simple:
                    plural_base = k.removesuffix("_plural")
                    plurals_simple.remove(plural_base)
                    plurals = [k, f"{k}_plural"]
                elif "_" in k:
                    plural_base, digit = k.rsplit("_", 1)
                    if plural_base in plurals_multiple and digit.isdigit():
                        plurals_multiple.remove(plural_base)
                        plurals = [f"{plural_base}_{order}" for order in range(10)]
                if plurals:
                    sources = []
                    items = []
                    for key in plurals:
                        if key not in data:
                            break
                        processed.add(key)
                        sources.append(data[key])
                        items.append(key)
                    unit = self.UnitClass(multistring(sources), items)
                    newid = prev.extend("key", plural_base)
                    unit.set_unitid(newid)
                    yield unit
                    continue

                yield from self._extract_units(
                    v, stop, prev.extend("key", k), k, None, data
                )
        else:
            yield from super()._extract_units(
                data, stop, prev, name_node, name_last_node, last_node
            )


class I18NextV4Unit(I18NextUnit):
    """
    A i18next v4 format, JSON with plurals.

    See https://www.i18next.com/
    """

    @staticmethod
    def _is_valid_suffix(suffix: str) -> bool:
        return suffix in cldr_plural_categories

    def _get_plural_labels(self, count):
        base_name = self._get_base_name()
        plural_tags = self._store.get_plural_tags(self.target)  # ty:ignore[unresolved-attribute]
        return [f"{base_name}_{plural_tags[i]}" for i in range(count)]


class I18NextV4File(JsonNestedFile):
    """
    A i18next v4 format, this is nested JSON with several additions.

    See https://www.i18next.com/
    """

    UnitClass = I18NextV4Unit

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
        plural_tags=None,
    ):
        if prev is None:
            prev = self.UnitClass.IdClass([])
        if plural_tags is None:
            plural_tags = self.get_plural_tags()
        if isinstance(data, dict):
            processed_plural_bases = set()

            for k, v in data.items():
                suffix = ""
                plural_base = ""

                if "_" in k:
                    plural_base, suffix = k.rsplit("_", 1)

                if suffix in cldr_plural_categories:
                    if plural_base in processed_plural_bases:
                        continue
                    processed_plural_bases.add(plural_base)
                    items = [f"{plural_base}_{suffix}" for suffix in plural_tags]
                    sources = [data.get(key, "") for key in items]

                    unit = self.UnitClass(multistring(sources), items)
                    newid = prev.extend("key", plural_base)
                    unit.set_unitid(newid)
                    yield unit
                    continue

                yield from self._extract_units(
                    v,
                    stop,
                    prev.extend("key", k),
                    k,
                    None,
                    data,
                    plural_tags,
                )
        else:
            yield from super()._extract_units(
                data, stop, prev, name_node, name_last_node, last_node
            )


class FlatI18NextV4Unit(I18NextV4Unit):
    IdClass = FlatUnitId


class FlatI18NextV4File(I18NextV4File):
    """
    Flat json file with support of i18next v4 format plurals.

    See https://www.i18next.com/
    """

    UnitClass = FlatI18NextV4Unit


class GoTextUnitId(base.UnitId):
    """Preserves id as stored in the JSON file."""

    def __str__(self) -> str:
        return str(self.parts)

    def extend(self, key, value) -> Never:
        raise ValueError("Extend is not supported")

    @classmethod
    def from_string(cls, text):
        return cls(text)


class GoTextJsonUnit(BaseJsonUnit):
    ID_FORMAT = "{}"
    IdClass = GoTextUnitId

    def __init__(
        self,
        source=None,
        item=None,
        notes=None,
        placeholders=None,
        comment=None,
        message=None,
        meaning=None,
        key=None,
        fuzzy=None,
        position=None,
        **kwargs,
    ) -> None:
        super().__init__(source, item, notes, placeholders)
        self.comment = comment
        self.message = message
        self.meaning = meaning
        self.key = key
        self.fuzzy = fuzzy
        self.position = position

    def getvalue(self):
        target = self.target
        if isinstance(target, multistring):
            plural_tags = self._store.get_plural_tags(target)  # ty:ignore[unresolved-attribute]
            strings = self.sync_plural_count(target, plural_tags)
            target = {
                "select": {
                    "feature": "plural",
                }
            }
            if self.placeholders:
                target["select"]["arg"] = self.placeholders[0]["id"]
            target["select"]["cases"] = {
                plural: {"msg": strings[offset]}
                for offset, plural in enumerate(plural_tags)
            }  # ty:ignore[invalid-assignment]
        value = {"id": self._unitid.parts if self._unitid else self.getid()}
        if self.message:
            value["message"] = self.message
        if self.notes:
            value["translatorComment"] = self.notes
        if self.comment:
            value["comment"] = self.comment
        if self.key:
            value["key"] = self.key
        if self.fuzzy:
            value["fuzzy"] = self.fuzzy
        if self.position:
            value["position"] = self.position
        value["translation"] = target
        if self.placeholders:
            value["placeholders"] = self.placeholders
        return value

    def setid(self, value, unitid=None) -> None:
        if unitid is None:
            unitid = self.IdClass(value)
        # Skip BaseJsonUnit.setid override
        super(BaseJsonUnit, self).setid(str(unitid), unitid)


class GoTextJsonFile(JsonFile):
    """
    gotext JSON file.

    See following URLs for doc:

    https://pkg.go.dev/golang.org/x/text/cmd/gotext
    https://github.com/golang/text/tree/master/cmd/gotext/examples/extract/locales/en-US
    """

    UnitClass = GoTextJsonUnit

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        def _get_msg(cases, key):
            value = cases.get(key, None)
            if isinstance(value, dict):
                value = value.get("msg")
            # Direct string value and None as fallback
            return require_json_string(value, f"gotext plural case {key!r}")

        if prev is None:
            data = require_json_dict(data, "gotext")
            lang = data.get("language")
            if lang is not None:
                self.settargetlanguage(require_json_string(lang, "language"))
        messages = data.get("messages", [])
        if not isinstance(messages, list):
            raise base.ParseError("gotext messages do not contain array")
        for value in messages:
            if not isinstance(value, dict):
                raise base.ParseError(
                    f"gotext message does not contain object: {value!r}"
                )
            translation = value.get("translation", "")
            if isinstance(translation, dict):
                select = translation.get("select", {})
                if not isinstance(select, dict):
                    raise base.ParseError(
                        f"gotext translation select does not contain object: {select!r}"
                    )
                cases = select.get("cases", {})
                if not isinstance(cases, dict):
                    raise base.ParseError(
                        f"gotext translation cases do not contain object: {cases!r}"
                    )
                # Ordered list of plurals
                translation = multistring(
                    [
                        _get_msg(cases, key)
                        for key in cldr_plural_categories
                        if key in cases
                    ]
                )
            elif translation is not None and not isinstance(translation, (str, int)):
                raise base.ParseError(
                    f"gotext translation does not contain string: {translation!r}"
                )
            unit = self.UnitClass(
                source=translation,
                item=value.get("id", ""),
                notes=value.get("translatorComment", ""),
                placeholders=value.get("placeholders", []),
                comment=value.get("comment", None),
                message=value.get("message", None),
                meaning=value.get("meaning", None),
                key=value.get("key", None),
                fuzzy=value.get("fuzzy", None),
                position=value.get("position", None),
            )
            unit.setid(value.get("id", ""))
            yield unit

    def serialize(self, out) -> None:
        units = [unit.getvalue() for unit in self.units]
        file = {
            "language": self.gettargetlanguage(),
            "messages": units,
        }
        out.write(json.dumps(file, **self.dump_args).encode(self.encoding))
        out.write(b"\n")


class GoI18NJsonUnit(FlatJsonUnit):
    ID_FORMAT = "{}"

    def getvalue(self):
        target = self.target
        if isinstance(target, multistring):
            plural_tags = self._store.get_plural_tags(target)  # ty:ignore[unresolved-attribute]
            strings = self.sync_plural_count(target, plural_tags)
            target = {
                plural: strings[offset] for offset, plural in enumerate(plural_tags)
            }
        value = {"id": self.getid()}
        if self.notes:
            value["description"] = self.notes
        value["translation"] = target
        return value


class GoI18NJsonFile(JsonFile):
    """
    go-i18n JSON file.

    See following URLs for doc:

    https://github.com/nicksnyder/go-i18n/tree/v1
    https://pkg.go.dev/github.com/nicksnyder/go-i18n
    """

    UnitClass = GoI18NJsonUnit

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        for value in require_json_list(data, "go-i18n"):
            if not isinstance(value, dict):
                raise base.ParseError(
                    f"go-i18n entry does not contain object: {value!r}"
                )
            item = value.get("id", "")
            translation = value.get("translation", "")
            if isinstance(translation, dict):
                # Ordered list of plurals
                try:
                    translation = multistring(
                        [
                            require_json_string(
                                translation.get(key), f"Key {item!r} plural {key!r}"
                            )
                            for key in cldr_plural_categories
                            if key in translation
                        ]
                    )
                except (TypeError, ValueError) as error:
                    raise base.ParseError(
                        f'"{item}" is an object but does not contain plurals. Maybe this is not a go-i18n JSON file?'
                    ) from error
            elif translation is not None and not isinstance(translation, (str, int)):
                raise base.ParseError(
                    f"go-i18n translation does not contain string: {translation!r}"
                )
            description = value.get("description", "")
            description = require_json_string(description, "description")
            unit = self.UnitClass(
                translation,
                item,
                description,
            )
            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
            yield unit

    def serialize(self, out) -> None:
        units = [unit.getvalue() for unit in self.units]
        out.write(json.dumps(units, **self.dump_args).encode(self.encoding))
        out.write(b"\n")


class GoI18NV2JsonUnit(FlatJsonUnit):
    ID_FORMAT = "{}"

    def converttarget(self):
        # Special handling of single translations
        if not isinstance(self.target, multistring) or len(self.target.strings) == 1:
            if self.notes:
                return {"description": self.notes, "other": self.target}
            return self.target

        target = {}
        if self.notes:
            target["description"] = self.notes

        plural_tags = self._store.get_plural_tags(self.target)  # ty:ignore[unresolved-attribute]
        strings = self.sync_plural_count(self.target, plural_tags)
        for offset, plural in enumerate(plural_tags):
            target[plural] = strings[offset]

        return target


class GoI18NV2JsonFile(JsonFile):
    """
    go-i18n v2 JSON file.

    See following URLs for doc:

    https://github.com/nicksnyder/go-i18n
    https://pkg.go.dev/github.com/nicksnyder/go-i18n/v2
    """

    UnitClass = GoI18NV2JsonUnit

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        for id, value in require_json_dict(data, "go-i18n v2").items():
            if isinstance(value, str):
                unit = self.UnitClass(value, id)
            else:
                if not isinstance(value, dict):
                    raise base.ParseError(
                        f"Key {id!r} does not contain string or object: {value!r}"
                    )
                try:
                    translation = multistring(
                        [
                            require_json_string(
                                value.get(key), f"Key {id!r} plural {key!r}"
                            )
                            for key in cldr_plural_categories
                            if key in value
                        ]
                    )
                except (TypeError, ValueError) as error:
                    raise base.ParseError(
                        f'"{id}" is an object but does not contain plurals. Maybe this is not a go-i18n v2 JSON file?'
                    ) from error
                description = value.get("description", "")
                description = require_json_string(
                    description, f"Key {id!r} description"
                )
                unit = self.UnitClass(
                    translation,
                    id,
                    description,
                )
            unit.setid(id, unitid=self.UnitClass.IdClass.from_key(id))
            yield unit


class ARBJsonUnit(base.MetadataTranslationUnit, FlatJsonUnit):
    ID_FORMAT = "{}"

    def __init__(
        self,
        source=None,
        item=None,
        notes=None,
        placeholders=None,
        metadata=None,
        **kwargs,
    ) -> None:
        # Pass metadata to MetadataTranslationUnit via super().__init__
        super().__init__(
            source=source,
            item=item,
            notes=notes,
            placeholders=placeholders,
            metadata=metadata,
            **kwargs,
        )

    def storevalues(self, output) -> None:
        if self.notes:
            self.metadata["description"] = self.notes
        identifier = self.getid()
        if identifier == "@":
            for key, value in self.metadata.items():
                self.storevalue(output, value, override_key=key)
        else:
            self.storevalue(output, self.target, override_key=identifier)
            if self.metadata:
                self.storevalue(output, self.metadata, override_key=f"@{identifier}")

    def isheader(self):
        return self._id == "@"


class ARBJsonFile(JsonFile):
    """
    ARB JSON file.

    See following URLs for doc:

    https://github.com/google/app-resource-bundle/wiki/ApplicationResourceBundleSpecification
    https://docs.flutter.dev/development/accessibility-and-localization/internationalization#dart-tools
    """

    UnitClass = ARBJsonUnit

    def __init__(self, inputfile=None, filter=None, **kwargs) -> None:
        super().__init__(inputfile, filter, **kwargs)
        self.dump_args = {
            "separators": (",", ": "),
            "indent": 2,
            "ensure_ascii": False,
        }

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        data = require_json_dict(data, "ARB")
        # Extract metadata as header
        metadata = {key: value for key, value in data.items() if key.startswith("@@")}
        if metadata:
            unit = self.UnitClass(metadata=metadata)
            unit.setid("@")
            yield unit

        for item, value in data.items():
            if item.startswith("@"):
                continue
            if not isinstance(value, (str, int)):
                raise base.ParseError(
                    f"Key {item!r} does not contain string: {value!r}"
                )
            metadata = data.get(f"@{item}", {})
            if not isinstance(metadata, dict):
                raise base.ParseError(
                    f"Key '@{item}' does not contain object: {metadata!r}"
                )
            description = metadata.get("description", "")
            description = require_json_string(description, f"Key '@{item}' description")
            placeholders = metadata.get("placeholders", None)
            if placeholders is not None and not isinstance(placeholders, dict):
                raise base.ParseError(
                    f"Key '@{item}' placeholders do not contain object: "
                    f"{placeholders!r}"
                )
            unit = self.UnitClass(
                value,
                item,
                description,
                placeholders,
                metadata=metadata,
            )
            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
            yield unit


class FormatJSJsonUnit(FlatJsonUnit):
    def storevalues(self, output) -> None:
        value = {"defaultMessage": self.target}
        if self.notes:
            value["description"] = self.notes
        self.storevalue(output, value)


class FormatJSJsonFile(JsonFile):
    """
    FormatJS JSON file.

    See following URLs for doc:

    https://formatjs.github.io/docs/getting-started/message-extraction/
    """

    UnitClass = FormatJSJsonUnit

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        for item, value in require_json_dict(data, "FormatJS").items():
            if not isinstance(value, dict):
                raise base.ParseError(
                    f"Not a valid FormatJS file should be a object: {value!r}"
                )
            default_message = value.get("defaultMessage", "")
            default_message = require_json_string(
                default_message, f"Key {item!r} defaultMessage"
            )
            description = value.get("description", "")
            if not isinstance(description, (str, dict)):
                raise base.ParseError(
                    f"Key {item!r} description does not contain string or object: "
                    f"{description!r}"
                )
            unit = self.UnitClass(
                default_message,
                item,
                description,
            )
            unit.setid(item, unitid=self.UnitClass.IdClass.from_key(item))
            yield unit


class NextcloudJsonUnit(FlatJsonUnit):
    """A Nextcloud JSON entry."""

    ID_FORMAT = "{}"

    def converttarget(self) -> Any:
        """Convert target to appropriate format for Nextcloud JSON."""
        if isinstance(self.target, multistring):
            # Return as array for plurals
            return list(self.target.strings)
        return self.target

    @property
    def source(self):
        return self.getid()

    @source.setter
    def source(self, source) -> None:
        self.setid(source)


class NextcloudJsonFile(JsonFile[NextcloudJsonUnit]):
    """
    Nextcloud JSON file.

    Nextcloud apps use a JSON format with translations wrapped in a
    "translations" key. Plurals follow gettext conventions with keys like
    ``_%n singular_::_%n plural_`` and array values.

    See:
    https://docs.nextcloud.com/server/stable/developer_manual/basics/translations.html
    https://github.com/nextcloud-libraries/nextcloud-l10n/
    """

    UnitClass = NextcloudJsonUnit

    def __init__(
        self,
        inputfile: str | bytes | TextIO | BinaryIO | None = None,
        filter: Any = None,
        **kwargs,
    ) -> None:
        """Construct a Nextcloud JSON file."""
        super().__init__(inputfile, filter, **kwargs)
        # Store top-level elements outside 'translations' for preservation
        self._metadata: dict[str, Any] = {}

    def _extract_units(
        self,
        data: Any,
        stop: list[str] | None = None,
        prev: Any = None,
        name_node: str | int | None = None,
        name_last_node: str | int | None = None,
        last_node: dict | list | None = None,
    ) -> Generator[NextcloudJsonUnit]:
        """Extract units from the translations key only."""
        data = require_json_dict(data, "Nextcloud")
        # Store metadata (everything outside 'translations')
        for key, value in data.items():
            if key != "translations":
                self._metadata[key] = value

        # Only parse the translations key
        translations = data.get("translations", {})
        if not isinstance(translations, dict):
            raise base.ParseError(
                f"Nextcloud translations do not contain object: {translations!r}"
            )

        for key, value in translations.items():
            # Check if this is a plural form (gettext style)
            if isinstance(value, list):
                for item in value:
                    require_json_string(item, f"Key {key!r} plural translation")
                # Plural form - array of translations
                unit = self.UnitClass(multistring(value), key)
            elif value is None:
                unit = self.UnitClass(value, key)
            else:
                value = require_json_string(value, f"Key {key!r}")
                # Simple string translation
                unit = self.UnitClass(value, key)

            unit.setid(key, unitid=self.UnitClass.IdClass.from_key(key))
            yield unit

    def serialize(self, out: IO[bytes]) -> None:
        """Serialize to Nextcloud JSON format."""
        # Build translations dictionary
        translations = {}
        for unit in self.units:
            key = unit.getid()
            value = unit.converttarget()
            translations[key] = value

        # Combine metadata and translations
        output = dict(self._metadata)
        output["translations"] = translations

        out.write(json.dumps(output, **self.dump_args).encode(self.encoding))
        out.write(b"\n")


class RESJSONUnit(base.MetadataTranslationUnit, FlatJsonUnit):
    """A RESJSON entry with metadata support."""

    ID_FORMAT = "{}"

    def __init__(
        self,
        source=None,
        item=None,
        notes=None,
        placeholders=None,
        metadata=None,
        **kwargs,
    ) -> None:
        # Pass metadata to MetadataTranslationUnit via super().__init__
        super().__init__(
            source=source,
            item=item,
            notes=notes,
            placeholders=placeholders,
            metadata=metadata,
            **kwargs,
        )

    def storevalues(self, output) -> None:
        # Sync notes to metadata
        if self.notes:
            self.metadata["comment"] = self.notes
        elif "comment" in self.metadata and not self.notes:
            # Remove comment if notes were cleared
            del self.metadata["comment"]
        if self._source:
            self.metadata["source"] = self._source

        identifier = self.getid()
        # Store the main value
        self.storevalue(output, self.target, override_key=identifier)
        # Store metadata with _KEY.suffix pattern
        for key, value in self.metadata.items():
            metadata_key = f"_{identifier}.{key}"
            self.storevalue(output, value, override_key=metadata_key)

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source) -> None:
        self._source = source

    def getcontext(self):
        return self.getid()


class RESJSONFile(JsonFile):
    """
    RESJSON (JavaScript Resource File) format.

    This format uses `_KEY.DATA` syntax to attach metadata to translation strings.

    See following URL for doc:

    https://docs.rws.com/en-US/sdl-passolo-help-785448/add-in-for-javascript-object-notation-json-file-format-types-410873
    """

    UnitClass = RESJSONUnit

    def _extract_units(
        self,
        data,
        stop=None,
        prev=None,
        name_node=None,
        name_last_node=None,
        last_node=None,
    ):
        data = require_json_dict(data, "RESJSON")
        # First pass: identify all actual keys (not metadata)
        key_map = {
            key: key.removeprefix("_").rsplit(".", 1)[0]
            if key.startswith("_") and "." in key[1:]
            else key
            for key in data
        }
        # Second pass: collect metadata for each actual key, preserving order
        metadata = defaultdict(dict)
        translations = defaultdict(str)
        for key, value in data.items():
            if not isinstance(value, (str, int)):
                raise base.ParseError(f"Key {key!r} does not contain string: {value!r}")
            actual_key = key_map[key]
            if actual_key == key:
                # Data
                translations[key] = value
            else:
                suffix = key.rsplit(".", 1)[1]
                metadata[actual_key][suffix] = value

        # Extract units
        processed = set()
        for key in key_map.values():
            if key in processed:
                continue
            processed.add(key)
            unit = self.UnitClass(
                metadata[key].get("source", ""),
                key,
                metadata[key].get("comment", ""),
                metadata[key].get("placeholders", None),
                metadata=metadata[key],
            )
            unit.target = translations[key]
            unit.setid(key, unitid=self.UnitClass.IdClass.from_key(key))
            yield unit



#
# Copyright 2005-2009 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""module for parsing TMX translation memory files."""

from lxml import etree

from translate import __version__
from translate.misc.xml_helpers import safely_set_text, setXMLlang
from translate.storage import lisa


class tmxunit(lisa.LISAunit):
    """A single unit in the TMX file."""

    rootNode = "tu"
    languageNode = "tuv"
    textNode = "seg"

    def createlanguageNode(self, lang, text, purpose):  # ty:ignore[invalid-method-override]
        """Returns a langset xml Element setup with given parameters."""
        langset = etree.Element(self.languageNode)
        setXMLlang(langset, lang)
        seg = etree.SubElement(langset, self.textNode)
        # implied by the standard:
        # setXMLspace(seg, "preserve")
        safely_set_text(seg, text)

        return langset

    def _insert_element_before(self, element: etree._Element, tag: str) -> None:
        """
        Insert an element before the first occurrence of the specified tag.

        According to TMX DTD, elements must follow this order: note, prop, tuv.
        This helper method finds the first child matching the tag name
        and inserts the element before it, or appends at the end if none found.

        :param element: The element to insert
        :param tag: Tag name to search for
        """
        needle = self.xmlelement.find(tag)
        if needle is None:
            self.xmlelement.append(element)
        else:
            self.xmlelement.insert(self.xmlelement.index(needle), element)

    def getid(self):
        """
        Returns the identifier for this unit. The optional tuid property is
        used if available, otherwise we inherit .getid(). Note that the tuid
        property is only mandated to be unique from TMX 2.0.
        """
        id = self.xmlelement.get("tuid", "")
        return id or super().getid()

    def istranslatable(self):
        return bool(self.source)

    def addnote(self, text, origin=None, position="append") -> None:
        """
        Add a note specifically in a "note" tag.

        The origin parameter is ignored
        """
        note = etree.Element(self.namespaced("note"))
        safely_set_text(note, text.strip())

        # According to TMX DTD, notes should come before prop and tuv elements
        # Try to insert before prop first, if not found try tuv
        if self.xmlelement.find(self.namespaced("prop")) is not None:
            self._insert_element_before(note, self.namespaced("prop"))
        else:
            self._insert_element_before(note, self.namespaced(self.languageNode))

    def _getnotelist(self, origin: str | None = None) -> list[str]:
        """
        Returns the text from notes.

        :param origin: Ignored
        :return: The text from notes
        """
        note_nodes = self.xmlelement.iterdescendants(self.namespaced("note"))
        return [lisa.getText(note) for note in note_nodes]

    def getnotes(self, origin=None):
        return "\n".join(self._getnotelist(origin=origin))

    def removenotes(self, origin=None) -> None:
        """Remove all the translator notes."""
        notes = self.xmlelement.iterdescendants(self.namespaced("note"))
        for note in notes:
            self.xmlelement.remove(note)

    def adderror(self, errorname, errortext) -> None:
        """Adds an error message to this unit."""
        # TODO: consider factoring out: some duplication between XLIFF and TMX
        text = errorname
        if errortext:
            text += f": {errortext}"
        self.addnote(text, origin="pofilter")

    def geterrors(self):
        """Get all error messages."""
        # TODO: consider factoring out: some duplication between XLIFF and TMX
        notelist = self._getnotelist(origin="pofilter")
        errordict = {}
        for note in notelist:
            errorname, errortext = note.split(": ")
            errordict[errorname] = errortext
        return errordict

    def setcontext(self, context) -> None:
        context_prop = self.xmlelement.find(
            f"{self.namespaced('prop')}[@type='x-context']"
        )
        if context_prop is None:
            context_prop = etree.Element(self.namespaced("prop"))
            context_prop.set("type", "x-context")

            # According to TMX DTD, prop elements come after notes but before tuv elements
            self._insert_element_before(
                context_prop, self.namespaced(self.languageNode)
            )
        safely_set_text(context_prop, context)

    def getcontext(self):
        context_prop = self.xmlelement.find(
            f"{self.namespaced('prop')}[@type='x-context']"
        )
        if context_prop is not None and context_prop.text is not None:
            return context_prop.text
        return ""


class tmxfile(lisa.LISAfile):
    """Class representing a TMX file store."""

    UnitClass = tmxunit
    Name = "TMX Translation Memory"
    Mimetypes = ["application/x-tmx"]
    Extensions = ["tmx"]
    rootNode = "tmx"
    bodyNode = "body"
    XMLskeleton = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE tmx SYSTEM "tmx14.dtd">
<tmx version="1.4">
<header></header>
<body></body>
</tmx>"""

    def addheader(self) -> None:
        headernode = next(
            self.document.getroot().iterchildren(self.namespaced("header"))
        )
        headernode.set("creationtool", "Translate Toolkit")
        headernode.set("creationtoolversion", __version__.sver)
        headernode.set("segtype", "sentence")
        headernode.set("o-tmf", "UTF-8")
        headernode.set("adminlang", "en")
        # TODO: consider adminlang. Used for notes, etc. Possibly same as
        # targetlanguage
        headernode.set("srclang", self.sourcelanguage or "en")
        headernode.set("datatype", "PlainText")
        # headernode.set("creationdate", "YYYYMMDDTHHMMSSZ"
        # headernode.set("creationid", "CodeSyntax"

    def addtranslation(
        self, source, srclang, translation, translang, comment=None, context=None
    ) -> None:
        """Addtranslation method for testing old unit tests."""
        unit = self.addsourceunit(source)
        unit.target = translation
        if comment is not None and len(comment) > 0:
            unit.addnote(comment)
        if context is not None and len(context) > 0:
            unit.setcontext(context)

        tuvs = unit.xmlelement.iterdescendants(self.namespaced("tuv"))
        setXMLlang(next(tuvs), srclang)
        setXMLlang(next(tuvs), translang)

    def translate(self, sourcetext, sourcelang=None, targetlang=None):  # ty:ignore[invalid-method-override]
        """Method to test old unit tests."""
        return getattr(self.findunit(sourcetext), "target", None)



#
# Copyright 2014 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# translate is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <http://www.gnu.org/licenses/>.

import warnings
from functools import wraps


def deprecated(message=""):
    """
    Decorator that marks functions and methods as deprecated.

    A warning will be emitted when the function or method is used. If a custom
    message is provided, it will be shown after the default warning message.
    """

    def inner_render(func):
        @wraps(func)
        def new_func(*args, **kwargs):
            msg = message  # Hack to avoid UnboundLocalError.
            if msg:
                msg = f"\n{msg}"
            func_code = func.__code__
            warnings.warn_explicit(
                f"Call to deprecated function {func.__name__}.{msg}",
                category=DeprecationWarning,
                filename=func_code.co_filename,
                lineno=func_code.co_firstlineno + 1,
            )
            return func(*args, **kwargs)

        return new_func

    return inner_render



#
# Copyright 2002-2011 Zuza Software Foundation
# Copyright 2013 F Wolff
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Grep XLIFF, Gettext PO and TMX localization files.

Matches are output to snippet files of the same type which can then be reviewed
and later merged using :doc:`pomerge </commands/pomerge>`.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/pogrep.html
for examples and usage instructions.
"""

import locale
import re

from translate.lang import data
from translate.misc import optrecurse
from translate.misc.multistring import multistring
from translate.storage import factory
from translate.storage.poheader import poheader


class GrepMatch:
    """Just a small data structure that represents a search match."""

    # INITIALIZERS #
    def __init__(self, unit, part="target", part_n=0, start=0, end=0) -> None:
        self.unit = unit
        self.part = part
        self.part_n = part_n
        self.start = start
        self.end = end

    # ACCESSORS #
    def get_getter(self):
        if self.part == "target":
            if self.unit.hasplural():
                return lambda: self.unit.target.strings[self.part_n]
            return lambda: self.unit.target
        if self.part == "source":
            if self.unit.hasplural():
                return lambda: self.unit.source.strings[self.part_n]
            return lambda: self.unit.source
        if self.part == "notes":

            def getter():
                return self.unit.getnotes()[self.part_n]

            return getter
        if self.part == "locations":

            def getter():  # pylint: disable=function-redefined
                return self.unit.getlocations()[self.part_n]

            return getter
        raise TypeError(f"Unsupported part: {self.part}")

    def get_setter(self):
        if self.part == "target":
            if self.unit.hasplural():

                def setter(value) -> None:
                    strings = self.unit.target.strings
                    strings[self.part_n] = value
                    self.unit.target = strings

            else:

                def setter(value) -> None:
                    self.unit.target = value

            return setter
        raise TypeError(f"Unsupported part: {self.part}")

    # SPECIAL METHODS #
    def __str__(self) -> str:
        start, end = self.start, self.end
        start = max(start, 3)
        end = min(end, len(self.get_getter()()) - 3)
        matchpart = self.get_getter()()[start - 2 : end + 2]
        return f'<GrepMatch "{matchpart}" part={self.part}[{self.part_n}] start={self.start} end={self.end}>'

    def __repr__(self) -> str:
        return str(self)


def real_index(string, nfc_index):
    """
    Calculate the real index in the unnormalized string that corresponds to
    the index nfc_index in the normalized string.
    """
    max_length = len(string)
    lower = 0
    # The extra position is a sentinel for normalized indexes at or beyond the
    # end of the string. It also handles characters whose NFC form expands.
    upper = max_length + 1

    while lower < upper:
        length = (lower + upper) // 2
        if length > max_length or len(data.normalize(string[:length])) > nfc_index:
            upper = length
        else:
            lower = length + 1

    return min(lower - 1, max_length)


def find_matches(unit, part, strings, re_search):
    """Return the GrepFilter objects where re_search matches in strings."""
    matches = []
    for n, string in enumerate(strings):
        if not string:
            continue
        normalized = data.normalize(string)
        index_func = (lambda s, i: i) if normalized == string else real_index
        for matchobj in re_search.finditer(normalized):
            start = index_func(string, matchobj.start())
            end = index_func(string, matchobj.end())
            matches.append(GrepMatch(unit, part=part, part_n=n, start=start, end=end))
    return matches


class GrepFilter:
    def __init__(
        self,
        searchstring,
        searchparts,
        ignorecase=False,
        useregexp=False,
        invertmatch=False,
        keeptranslations=False,
        accelchar=None,
        encoding="utf-8",
        max_matches=0,
    ) -> None:
        """Builds a checkfilter using the given checker."""
        if isinstance(searchstring, str):
            self.searchstring = searchstring
        else:
            self.searchstring = searchstring.decode(encoding)
        self.searchstring = data.normalize(self.searchstring)
        if searchparts:
            # For now we still support the old terminology, except for the old 'source'
            # which has a new meaning now.
            self.search_source = ("source" in searchparts) or ("msgid" in searchparts)
            self.search_target = ("target" in searchparts) or ("msgstr" in searchparts)
            self.search_notes = ("notes" in searchparts) or ("comment" in searchparts)
            self.search_locations = "locations" in searchparts
        else:
            self.search_source = True
            self.search_target = True
            self.search_notes = False
            self.search_locations = False
        self.ignorecase = ignorecase
        if self.ignorecase:
            self.searchstring = self.searchstring.lower()
        self.useregexp = useregexp
        if self.useregexp:
            self.searchpattern = re.compile(self.searchstring)
        self.invertmatch = invertmatch
        self.keeptranslations = keeptranslations
        self.accelchar = accelchar
        self.max_matches = max_matches

    def matches(self, teststr):
        if teststr is None:
            return False
        teststr = data.normalize(teststr)
        if self.ignorecase:
            teststr = teststr.lower()
        if self.accelchar:
            teststr = re.sub(self.accelchar + self.accelchar, "#", teststr)
            teststr = re.sub(self.accelchar, "", teststr)
        if self.useregexp:
            found = self.searchpattern.search(teststr)
        else:
            found = teststr.find(self.searchstring) != -1
        if self.invertmatch:
            found = not found
        return found

    def filterunit(self, unit):
        """Runs filters on an element."""
        if unit.isheader():
            return True

        if self.keeptranslations and unit.target:
            return True

        if self.search_source:
            if isinstance(unit.source, multistring):
                strings = unit.source.strings
            else:
                strings = [unit.source]
            for string in strings:
                if self.matches(string):
                    return True

        if self.search_target:
            if isinstance(unit.target, multistring):
                strings = unit.target.strings
            else:
                strings = [unit.target]
            for string in strings:
                if self.matches(string):
                    return True

        if self.search_notes and self.matches(unit.getnotes()):
            return True
        return bool(
            self.search_locations and self.matches(" ".join(unit.getlocations()))
        )

    def filterfile(self, thefile):
        """Runs filters on a translation file object."""
        thenewfile = type(thefile)()
        thenewfile.setsourcelanguage(thefile.sourcelanguage)
        thenewfile.settargetlanguage(thefile.targetlanguage)
        for unit in thefile.units:
            if self.filterunit(unit):
                thenewfile.addunit(unit)

        if isinstance(thenewfile, poheader):
            thenewfile.updateheader(add=True, **thefile.parseheader())
        return thenewfile

    def getmatches(self, units):
        if not self.searchstring:
            return [], []

        searchstring = self.searchstring
        flags = re.LOCALE | re.MULTILINE | re.UNICODE

        if self.ignorecase:
            flags |= re.IGNORECASE
        if not self.useregexp:
            searchstring = re.escape(searchstring)
        self.re_search = re.compile(f"({searchstring})", flags)

        matches = []
        indexes = []

        for index, unit in enumerate(units):
            old_length = len(matches)

            if self.search_target:
                targets = unit.target.strings if unit.hasplural() else [unit.target]
                matches.extend(find_matches(unit, "target", targets, self.re_search))
            if self.search_source:
                sources = unit.source.strings if unit.hasplural() else [unit.source]
                matches.extend(find_matches(unit, "source", sources, self.re_search))
            if self.search_notes:
                matches.extend(
                    find_matches(unit, "notes", unit.getnotes(), self.re_search)
                )

            if self.search_locations:
                matches.extend(
                    find_matches(unit, "locations", unit.getlocations(), self.re_search)
                )

            # A search for a single letter or an all-inclusive regular
            # expression could give enough results to cause performance
            # problems. The answer is probably not very useful at this scale.
            if self.max_matches and len(matches) > self.max_matches:
                raise ValueError("Too many matches found")

            if len(matches) > old_length:
                old_length = len(matches)
                indexes.append(index)

        return matches, indexes


class GrepOptionParser(optrecurse.RecursiveOptionParser):
    """a specialized Option Parser for the grep tool..."""

    def parse_args(self, args=None, values=None):
        """Parses the command line options, handling implicit input/output args."""
        (options, args) = optrecurse.optparse.OptionParser.parse_args(
            self, args, values
        )
        # some intelligence as to what reasonable people might give on the command line
        if args:
            options.searchstring = args[0]
            args = args[1:]
        else:
            self.error("At least one argument must be given for the search string")
        if args and not options.input:
            if not options.output:
                options.input = args[:-1]
                args = args[-1:]
            else:
                options.input = args
                args = []
        if args and not options.output:
            options.output = args[-1]
            args = args[:-1]
        if args:
            self.error(
                "You have used an invalid combination of --input, --output and freestanding args"
            )
        if isinstance(options.input, list) and len(options.input) == 1:
            options.input = options.input[0]
        return (options, args)

    def set_usage(self, usage=None) -> None:
        """Sets the usage string - if usage not given, uses getusagestring for each option."""
        if usage is None:
            self.usage = f"%prog searchstring {' '.join(self.getusagestring(option) for option in self.option_list)}"
        else:
            super().set_usage(usage)

    def run(self) -> None:
        """Parses the arguments, and runs recursiveprocess with the resulting options."""
        options, _args = self.parse_args()
        options.checkfilter = GrepFilter(
            options.searchstring,
            options.searchparts,
            options.ignorecase,
            options.useregexp,
            options.invertmatch,
            options.keeptranslations,
            options.accelchar,
            locale.getpreferredencoding(),
        )
        self.recursiveprocess(options)


def rungrep(inputfile, outputfile, templatefile, checkfilter) -> bool:
    """Reads in inputfile, filters using checkfilter, writes to outputfile."""
    fromfile = factory.getobject(inputfile)
    tofile = checkfilter.filterfile(fromfile)
    if tofile.isempty():
        return False
    tofile.serialize(outputfile)
    return True


def cmdlineparser():
    formats = {
        "po": ("po", rungrep),
        "pot": ("pot", rungrep),
        "mo": ("mo", rungrep),
        "gmo": ("gmo", rungrep),
        "tmx": ("tmx", rungrep),
        "xliff": ("xliff", rungrep),
        "xlf": ("xlf", rungrep),
        "xlff": ("xlff", rungrep),
        None: ("po", rungrep),
    }
    parser = GrepOptionParser(formats)
    parser.add_option(
        "",
        "--search",
        dest="searchparts",
        action="append",
        type="choice",
        choices=[
            "source",
            "target",
            "notes",
            "locations",
            "msgid",
            "msgstr",
            "comment",
        ],
        metavar="SEARCHPARTS",
        help="searches the given parts (source, target, notes and locations)",
    )
    parser.add_option(
        "-I",
        "--ignore-case",
        dest="ignorecase",
        action="store_true",
        default=False,
        help="ignore case distinctions",
    )
    parser.add_option(
        "-e",
        "--regexp",
        dest="useregexp",
        action="store_true",
        default=False,
        help="use regular expression matching",
    )
    parser.add_option(
        "-v",
        "--invert-match",
        dest="invertmatch",
        action="store_true",
        default=False,
        help="select non-matching lines",
    )
    parser.add_option(
        "",
        "--accelerator",
        dest="accelchar",
        action="store",
        type="choice",
        choices=["&", "_", "~"],
        metavar="ACCELERATOR",
        help="ignores the given accelerator when matching",
    )
    parser.add_option(
        "-k",
        "--keep-translations",
        dest="keeptranslations",
        action="store_true",
        default=False,
        help="always extract units with translations",
    )
    parser.set_usage()
    parser.passthrough.append("checkfilter")
    parser.description = __doc__
    return parser


def main() -> None:
    parser = cmdlineparser()
    parser.run()


if __name__ == "__main__":
    main()



#
# Copyright 2008-2017 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""This file contains the version of the Translate Toolkit."""

ver = (3, 19, 14)
"""Machine readable version number. Used by tools that need to adjust code
paths based on a Translate Toolkit release number."""

build = ver[0] * 10000 + ver[1] * 100 + ver[2]
"""The build number is used by external users of the Translate Toolkit to
trigger refreshes.  Thus increase the build number whenever changes are made to
code touching stats or quality checks.  An increased build number will force a
toolkit user, like Pootle, to regenerate its stored stats and check
results."""

sver = f"{ver[0]}.{ver[1]}.{ver[2]}"
"""Human readable version number. Used for version number display."""



# noqa: N999
#
# Copyright 2016 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

"""
This module represents the Brazilian Portuguese language.

.. seealso:: :wp:`Brazilian_Portuguese`
"""

from translate.lang.common import Common


class pt_BR(Common):
    """This class represents Brazilian Portuguese."""

    miscpunc = Common.miscpunc.replace("·", "")  # Middle dot is not valid.

    validaccel = "ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvxyz1234567890"



#
# Copyright 2008-2014 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager

from lxml import etree

from translate.misc.xml_helpers import parse_xml_file
from translate.storage.placeables import StringElem, xliff
from translate.storage.xliff import xliffunit
from translate.storage.xml_extract import misc, xpath_breadcrumb


class Translatable:
    """
    A node corresponds to a translatable element. A node may have children,
    which correspond to placeables.
    """

    def __init__(
        self, placeable_name, xpath, dom_node, source, is_inline=False
    ) -> None:
        self.placeable_name = placeable_name
        self.source = source
        self.xpath = xpath
        self.is_inline = is_inline
        self.dom_node = dom_node

    @property
    def placeables(self):
        return [
            placeable
            for placeable in self.source
            if isinstance(placeable, Translatable)
        ]

    @property
    def has_translatable_text(self):
        """
        Check if it contains any chunk of text with more than whitespace.

        If not, then there's nothing to translate.
        """
        return any(isinstance(chunk, str) and chunk.strip() for chunk in self.source)


def reduce_unit_tree(f, unit_node, *state):
    return misc.reduce_tree(
        f, unit_node, unit_node, lambda unit_node: unit_node.placeables, *state
    )


class ParseState:
    """
    Maintain constants and variables used during the walking of a DOM tree
    (via the function apply).
    """

    def __init__(
        self, no_translate_content_elements, inline_elements={}, nsmap={}
    ) -> None:
        self.no_translate_content_elements = no_translate_content_elements
        self.inline_elements = inline_elements
        self.is_inline = False
        self.xpath_breadcrumb = xpath_breadcrumb.XPathBreadcrumb()
        self.placeable_name = "<top-level>"
        self.nsmap = nsmap


def _process_placeable(dom_node, state):
    """
    Process current placeable.

    This returns all nested translatable content for this placeable as a single
    Translatable object, or just returns an empty Translatable object for this
    placeable if there is no nested translatable content.
    """
    placeable = find_translatable_dom_nodes(dom_node, state, process_translatable)

    if len(placeable) == 0:
        # There are no recognized child tags and thus no Translatable object is
        # returned. So create a Translatable with the name "placeable".
        return Translatable("placeable", state.xpath_breadcrumb.xpath, dom_node, [])
    if len(placeable) == 1:
        # The ideal situation: we got exactly one Translatable back when
        # processing this tree.
        return placeable[0]
    raise ValueError(
        "BUG: find_translatable_dom_nodes should never return "
        "more than a single Translatable object"
    )


def process_translatable(dom_node, state):
    """
    Process a translatable DOM node.

    Any translatable content present in a child node is treated as a placeable.
    """
    source = [str(dom_node.text or "")]

    # Append Translatable objects and unicode strings for the translatable
    # content for all the children.
    for child in dom_node:
        source.extend((_process_placeable(child, state), str(child.tail or "")))

    translatable = Translatable(
        state.placeable_name,
        state.xpath_breadcrumb.xpath,
        dom_node,
        source,
        state.is_inline,
    )
    return [translatable]


def _has_idml_translatable_content(dom_node):
    has_translatable_content = True
    if dom_node.tag == "ParagraphStyleRange":
        has_translatable_content = False

        for content_node in dom_node.findall(".//Content"):
            # Iterate over all the Content tags in this ParagraphStyleRange tag.
            if content_node.text is not None and content_node.text.strip():
                return True

            for child in content_node.iterdescendants():
                # The Content node can just have a child before any nested
                # text, and therefore its text is None, so we have to check if
                # it has children, and if any of its children has text.
                if (
                    not isinstance(child, etree._ProcessingInstruction)
                    and child.text is not None
                    and child.text.strip()
                ):
                    return True

                if child.tail is not None and child.tail.strip():
                    return True

    return has_translatable_content


def _retrieve_idml_placeables(dom_node, state):
    source = []
    for child in dom_node:
        if not isinstance(child, etree._Element):
            continue

        if isinstance(child, etree._ProcessingInstruction):
            # TODO this probably won't be using the right xpath.
            source.append(
                Translatable(
                    "placeable", state.xpath_breadcrumb.xpath, child, [], False
                )
            )

            if child.tail is not None and child.tail.strip():
                source.append(str(child.tail))

            continue

        namespace, tag = misc.parse_tag(child.tag)

        with parse_status_set(namespace, tag, state):
            # Ensure we extract all the tags below ParagraphStyleRange as
            # placeables, independently of them being translatable or not.
            # state.is_inline = True

            nested_stuff = []

            if child.text is not None and child.text.strip():
                nested_stuff = [str(child.text)]

            nested_stuff.extend(_retrieve_idml_placeables(child, state))

            source.append(
                Translatable(
                    "placeable",
                    state.xpath_breadcrumb.xpath,
                    child,
                    nested_stuff,
                    state.is_inline,
                )
            )

            if child.tail is not None and child.tail.strip():
                source.append(str(child.tail))

    return source


def process_idml_translatable(dom_node, state):
    if _has_idml_translatable_content(dom_node):
        source = _retrieve_idml_placeables(dom_node, state)

        translatable = Translatable(
            state.placeable_name,
            state.xpath_breadcrumb.xpath,
            dom_node,
            source,
            state.is_inline,
        )
        return [translatable]

    return []


def _process_children(dom_node, state, process_func):
    """
    Process an untranslatable DOM node.

    Since the node is untranslatable it just returns any translatable content
    present in its child nodes.
    """
    children = [
        find_translatable_dom_nodes(child, state, process_func) for child in dom_node
    ]

    # Flatten a list of lists into a list of elements
    children = [child for child_list in children for child in child_list]

    if len(children) > 1:
        _namespace, tag = misc.parse_tag(dom_node.tag)
        intermediate_translatable = Translatable(
            tag, state.xpath_breadcrumb.xpath, dom_node, children
        )
        return [intermediate_translatable]
    return children


def compact_tag(nsmap, namespace, tag) -> str:
    if namespace in nsmap:
        return f"{nsmap[namespace]}:{tag}"
    return f"{{{namespace}}}{tag}"


@contextmanager
def parse_status_set(namespace, tag, state):
    # Set XPath breadcrumb item for the current node.
    xpath_item = compact_tag(state.nsmap, namespace, tag)
    state.xpath_breadcrumb.start_tag(xpath_item)

    # Set the placeable name for the current node.
    old_placeable_name = state.placeable_name
    state.placeable_name = tag

    # Set the inline status for the current node.
    old_inline = state.is_inline
    state.is_inline = (namespace, tag) in state.inline_elements

    try:
        yield state
    finally:
        # Reset inline status, placeable name and XPath breadcrumb to the
        # previous values.
        state.is_inline = old_inline
        state.placeable_name = old_placeable_name
        state.xpath_breadcrumb.end_tag()


def find_translatable_dom_nodes(dom_node, state, process_func=process_translatable):
    # For now, we only want to deal with XML elements.
    # And we want to avoid processing instructions, which
    # are XML elements (in the inheritance hierarchy).
    if not isinstance(dom_node, etree._Element) or isinstance(
        dom_node, etree._ProcessingInstruction
    ):
        return []

    namespace, tag = misc.parse_tag(dom_node.tag)

    with parse_status_set(namespace, tag, state):
        if (namespace, tag) not in state.no_translate_content_elements:
            return process_func(dom_node, state)
        return _process_children(dom_node, state, process_func)


class IdMaker:
    def __init__(self) -> None:
        self._max_id = 0
        self._obj_id_map = {}

    def get_id(self, obj):
        if not self.has_id(obj):
            self._obj_id_map[obj] = self._max_id
            self._max_id += 1
        return self._obj_id_map[obj]

    def has_id(self, obj):
        return obj in self._obj_id_map


def _to_placeables(parent_translatable, translatable, id_maker):
    """
    Convert the translatable object to a list of strings and XLIFF
    placeables.
    """
    result = []
    for chunk in translatable.source:
        if isinstance(chunk, str):
            result.append(chunk)
        else:
            id = str(id_maker.get_id(chunk))
            if chunk.is_inline:
                sub = _to_placeables(parent_translatable, chunk, id_maker)
                result.append(xliff.G(id=id, sub=sub))
            else:
                result.append(xliff.X(id=id, xid=chunk.xpath))
    return result


def _make_store_adder(store):
    """
    Return a function which, when called with a Translatable will add
    a unit to 'store'. The placeables will be represented as strings according
    to 'placeable_quoter'.
    """
    id_maker = IdMaker()

    def add_translatable_to_store(parent_translatable, translatable) -> None:
        """
        Construct a new translation unit, set its source and location
        information and add it to 'store'.
        """
        unit = store.UnitClass("")
        unit.rich_source = [
            StringElem(_to_placeables(parent_translatable, translatable, id_maker))
        ]
        if isinstance(unit, xliffunit):
            unit.setid(translatable.xpath)
        else:
            unit.addlocation(translatable.xpath)
        store.addunit(unit)

    return add_translatable_to_store


def make_postore_adder(store, id_maker, filename):
    """
    Return a function which, when called with a Translatable will add
    a unit to 'store'. The placeables will be represented as strings according
    to 'placeable_quoter'.
    """

    def add_translatable_to_store(parent_translatable, translatable) -> None:
        """
        Construct a new translation unit, set its source and location
        information and add it to 'store'.
        """
        xliff_unit = xliffunit("")
        placeables = _to_placeables(parent_translatable, translatable, id_maker)
        xliff_unit.rich_source = [StringElem(placeables)]

        # Get the plain text for the unit source. The output is enclosed within
        # XLIFF source tags we don't want, so strip them.
        unit_source = etree.tostring(xliff_unit.source_dom, encoding="unicode")
        unit_source = unit_source[unit_source.find(">", 1) + 1 :]
        unit_source = unit_source[: unit_source.rfind("<", 1)]

        # Create the PO unit and add it to the PO store.
        po_unit = store.UnitClass(unit_source)
        po_unit.addlocation(translatable.xpath)
        po_unit.addlocation(filename)
        store.addunit(po_unit)

    return add_translatable_to_store


def _walk_idml_translatable_tree(
    translatables, store_adder, parent_translatable
) -> None:
    """
    Traverse all the found IDML translatables and add them to the Store.

    Inline translatables are not added to the Store.
    """
    for translatable in translatables:
        if translatable.dom_node.tag == "ParagraphStyleRange":
            store_adder(parent_translatable, translatable)
            continue

        new_parent_translatable = parent_translatable
        _walk_idml_translatable_tree(
            translatable.placeables, store_adder, new_parent_translatable
        )


def _walk_translatable_tree(
    translatables, store_adder, parent_translatable, stored_by_parent=False
) -> None:
    """
    Traverse all the found translatables and add them to the Store.

    Inline translatables are not added to the Store.
    """
    for translatable in translatables:
        store_here = translatable.has_translatable_text and (
            not translatable.is_inline or not stored_by_parent
        )
        if store_here:
            store_adder(parent_translatable, translatable)
        new_parent_translatable = parent_translatable
        _walk_translatable_tree(
            translatable.placeables,
            store_adder,
            new_parent_translatable,
            store_here or stored_by_parent,
        )


def reverse_map(a_map):
    return {value: key for key, value in a_map.items()}


def build_idml_store(odf_file, store, parse_state, store_adder=None):
    """Build a store for the given IDML file."""
    store_adder = store_adder or _make_store_adder(store)
    tree = parse_xml_file(odf_file)
    root = tree.getroot()
    parse_state.nsmap = reverse_map(root.nsmap)
    translatables = find_translatable_dom_nodes(
        root, parse_state, process_idml_translatable
    )
    _walk_idml_translatable_tree(translatables, store_adder, None)
    return tree


def build_store(odf_file, store, parse_state, store_adder=None):
    """Build a store for the given XML file."""
    store_adder = store_adder or _make_store_adder(store)
    tree = parse_xml_file(odf_file)
    root = tree.getroot()
    parse_state.nsmap = reverse_map(root.nsmap)
    translatables = find_translatable_dom_nodes(root, parse_state)
    _walk_translatable_tree(translatables, store_adder, None)
    return tree



#
# Copyright 2007-2008 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Class that manages iCalender files for translation.

iCalendar files follow the `RFC2445 <https://tools.ietf.org/html/rfc2445>`_
specification.

The iCalendar specification uses the following naming conventions:

- Component: an event, journal entry, timezone, etc
- Property: a property of a component: summary, description, start time, etc
- Attribute: an attribute of a property, e.g. language

The following are localisable in this implementation:

- VEVENT component: SUMMARY, DESCRIPTION, COMMENT and LOCATION properties

While other items could be localised this is not seen as important until use
cases arise.  In such a case simply adjusting the component.name and
property.name lists to include these will allow expanded localisation.

LANGUAGE Attribute
    While the iCalendar format allows items to have a language attribute this is
    not used. The reason being that for most of the items that we localise they
    are only allowed to occur zero or once.  Thus 'summary' would ideally
    be present in multiple languages in one file, the format does not allow
    such multiple entries.  This is unfortunate as it prevents the creation
    of a single multilingual iCalendar file.

Future Format Support
    As this format used `vobject <http://eventable.github.io/vobject/>`_
    which supports various formats including
    :wp:`vCard` it is possible to expand
    this format to understand those if needed.
"""

import re
from io import BytesIO

import vobject

from translate.storage import base

ICAL_UNIT_LOCATION_RE = re.compile(r"\[(?P<uid>.+)\](?P<property>.+)")


class icalunit(base.TranslationUnit):
    """An ical entry that is translatable."""

    def __init__(self, source=None, **kwargs) -> None:
        self.location = ""
        if source:
            self.source = source
        super().__init__(source)

    def addlocation(self, location) -> None:
        self.location = location

    def getlocations(self):
        return [self.location]


class icalfile(base.TranslationStore):
    """An ical file."""

    UnitClass = icalunit

    def __init__(self, inputfile=None, **kwargs) -> None:
        """Construct an ical file, optionally reading in from inputfile."""
        super().__init__(**kwargs)
        self.filename = ""
        self._icalfile = None
        if inputfile is not None:
            self.parse(inputfile)

    def serialize(self, out) -> None:
        outicalfile = self._icalfile
        for unit in self.units:
            for location in unit.getlocations():
                match = ICAL_UNIT_LOCATION_RE.match(location)
                for component in self._icalfile.components():  # ty:ignore[unresolved-attribute]
                    if component.name != "VEVENT":
                        continue
                    if component.uid.value != match.groupdict()["uid"]:  # ty:ignore[unresolved-attribute]
                        continue
                    for property in component.getChildren():
                        if property.name == match.groupdict()["property"]:  # ty:ignore[unresolved-attribute]
                            property.value = unit.target

        if outicalfile:
            outicalfile.serialize(out)

    def parse(self, input) -> None:  # ty:ignore[invalid-method-override]
        """Parse the given file or file source string."""
        if hasattr(input, "name"):
            self.filename = input.name
        elif not getattr(self, "filename", ""):
            self.filename = ""
        if isinstance(input, BytesIO):
            input = input.getvalue()
        elif hasattr(input, "read"):
            inisrc = input.read()
            input.close()
            input = inisrc
        if isinstance(input, bytes):
            input = input.decode(self.encoding)
        self._icalfile = next(vobject.readComponents(input))

        for component in self._icalfile.components():
            if component.name == "VEVENT":
                for property in component.getChildren():
                    if property.name in {
                        "SUMMARY",
                        "DESCRIPTION",
                        "COMMENT",
                        "LOCATION",
                    }:
                        newunit = self.addsourceunit(property.value)
                        newunit.addnote(f"Start date: {component.dtstart.value}")
                        newunit.addlocation(f"[{component.uid.value}]{property.name}")



#
# Copyright 2024 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

"""
Convert MDX files to Gettext PO localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/mdx2po.html
for examples and usage instructions.
"""

from __future__ import annotations

import functools

from translate.convert import convert
from translate.storage import mdxfile, po


class MDX2POOptionParser(convert.ConvertOptionParser, convert.DocpathMerger):
    def __init__(self) -> None:
        formats = {
            "mdx": ("po", self._extract_translation_units),
            None: ("po", self._extract_translation_units),
        }
        super().__init__(formats, usetemplates=True, usepots=True, description=__doc__)
        self.add_duplicates_option()
        self.add_multifile_option()
        self.add_option(
            "",
            "--no-code-blocks",
            action="store_false",
            dest="extract_code_blocks",
            default=True,
            help="do not extract code blocks for translation",
        )
        self.passthrough.append("extract_code_blocks")
        self.add_option(
            "",
            "--no-frontmatter",
            action="store_false",
            dest="extract_frontmatter",
            default=True,
            help="do not extract front matter for translation",
        )
        self.passthrough.append("extract_frontmatter")
        self.add_option(
            "",
            "--no-placeholders",
            action="store_true",
            dest="no_placeholders",
            default=False,
            help=(
                "render inline elements (links, images, autolinks, HTML spans) "
                "verbatim in translation units instead of replacing them with "
                "{n} placeholder markers"
            ),
        )
        self.passthrough.append("no_placeholders")

    def _extract_translation_units(
        self,
        inputfile,
        outputfile,
        templatefile,
        duplicatestyle: str,
        multifilestyle: str,
        extract_code_blocks: bool = True,
        extract_frontmatter: bool = True,
        no_placeholders: bool = False,
    ) -> int:
        if hasattr(self, "outputstore"):
            if templatefile is None:
                self._parse_and_extract(
                    inputfile,
                    self.outputstore,
                    extract_code_blocks=extract_code_blocks,
                    extract_frontmatter=extract_frontmatter,
                    no_placeholders=no_placeholders,
                )
            else:
                self._merge_with_template(
                    inputfile,
                    templatefile,
                    self.outputstore,
                    extract_code_blocks=extract_code_blocks,
                    extract_frontmatter=extract_frontmatter,
                    no_placeholders=no_placeholders,
                )
        else:
            store = po.pofile()
            if templatefile is None:
                self._parse_and_extract(
                    inputfile,
                    store,
                    extract_code_blocks=extract_code_blocks,
                    extract_frontmatter=extract_frontmatter,
                    no_placeholders=no_placeholders,
                )
            else:
                self._merge_with_template(
                    inputfile,
                    templatefile,
                    store,
                    extract_code_blocks=extract_code_blocks,
                    extract_frontmatter=extract_frontmatter,
                    no_placeholders=no_placeholders,
                )
            store.removeduplicates(duplicatestyle)
            store.serialize(outputfile)
        return 1

    @staticmethod
    def _parse_and_extract(
        inputfile,
        outputstore: po.pofile,
        *,
        extract_code_blocks: bool = True,
        extract_frontmatter: bool = True,
        no_placeholders: bool = False,
    ) -> None:
        """Extract translation units from an MDX file."""
        parser = mdxfile.MDXFile(
            inputfile=inputfile,
            extract_code_blocks=extract_code_blocks,
            extract_frontmatter=extract_frontmatter,
            no_placeholders=no_placeholders,
        )
        for tu in parser.units:
            storeunit = outputstore.addsourceunit(tu.source)
            storeunit.addlocations(tu.getlocations())

    def _merge_with_template(
        self,
        inputfile,
        templatefile,
        outputstore: po.pofile,
        *,
        extract_code_blocks: bool = True,
        extract_frontmatter: bool = True,
        no_placeholders: bool = False,
    ) -> None:
        """Merge translation from inputfile with source from templatefile."""
        store_class = functools.partial(
            mdxfile.MDXFile,
            extract_code_blocks=extract_code_blocks,
            extract_frontmatter=extract_frontmatter,
            no_placeholders=no_placeholders,
        )
        self.merge_stores_by_docpath(
            inputfile,
            templatefile,
            outputstore,
            store_class,
            filter_header=True,
        )

    def recursiveprocess(self, options) -> None:
        """Recurse through directories and process files. (override)."""
        if options.multifilestyle == "onefile":
            self.outputstore = po.pofile()
            super().recursiveprocess(options)
            if not self.outputstore.isempty():
                outputfile = super().openoutputfile(options, options.output)
                self.outputstore.removeduplicates(options.duplicatestyle)
                self.outputstore.serialize(outputfile)
                if options.output:
                    outputfile.close()
        else:
            super().recursiveprocess(options)

    def isrecursive(self, fileoption, filepurpose="input"):
        """Check if fileoption is a recursive file. (override)."""
        if hasattr(self, "outputstore") and filepurpose == "output":
            return True
        return super().isrecursive(fileoption, filepurpose=filepurpose)

    def checkoutputsubdir(self, options, subdir) -> None:
        """Check if subdir needs to be created. (override)."""
        if hasattr(self, "outputstore"):
            return
        super().checkoutputsubdir(options, subdir)

    def openoutputfile(self, options, fulloutputpath):
        """Open the output file. (override)."""
        if hasattr(self, "outputstore"):
            return None
        return super().openoutputfile(options, fulloutputpath)


def main(argv=None) -> None:
    parser = MDX2POOptionParser()
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2025 translate-toolkit contributors
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Module for parsing AsciiDoc files for translation.

The principles for extraction of translation units are as follows:

1. Extract all content relevant for translation, at the cost of also
   including some formatting.
2. One translation unit per paragraph.
3. Keep formatting out of the translation units as much as possible.
   Use placeholders {1}, {2}, ..., as needed for complex inline elements.
4. Support common AsciiDoc elements: headings, paragraphs, lists, etc.

White space within translation units is normalized.
"""

from __future__ import annotations

import re
from typing import Any

from translate.storage import base


class AsciiDocUnit(base.TranslationUnit):
    """A unit of translatable/localisable AsciiDoc content."""

    def __init__(self, source: str | None = None) -> None:
        super().__init__(source)
        self.locations: list[str] = []
        self._element_type: str = "paragraph"
        self._prefix: str = ""
        self._suffix: str = ""

    def addlocation(self, location: str) -> None:
        self.locations.append(location)

    def getlocations(self) -> list[str]:
        return self.locations

    def set_element_info(
        self, element_type: str, prefix: str = "", suffix: str = ""
    ) -> None:
        """Store element type and formatting information."""
        self._element_type = element_type
        self._prefix = prefix
        self._suffix = suffix


class AsciiDocHeaderUnit(AsciiDocUnit):
    def isheader(self) -> bool:
        """Return True to indicate this unit represents document header metadata."""
        return True


class AsciiDocFile(base.TranslationStore):
    UnitClass = AsciiDocUnit

    def __init__(self, inputfile=None, callback=None) -> None:
        """
        Construct a new object instance.

        :param inputfile: if specified, the content of this file is read and parsed.
        :param callback: a function which takes a chunk of untranslated content as
          input and returns the corresponding translated content. Defaults to
          a no-op.
        """
        base.TranslationStore.__init__(self)
        self.filename: str | None = getattr(inputfile, "name", None)
        self.callback = callback or self._dummy_callback
        self.filesrc: str = ""
        self._elements: list[
            dict[str, Any]
        ] = []  # Store parsed elements for reconstruction
        # Docpath tracking: hierarchical heading structure
        self._heading_stack: list[tuple[int, int]] = []  # (level, index) pairs
        self._section_counts: list[dict[str, int]] = [
            {}
        ]  # Stack of element counts per section
        if inputfile is not None:
            adoc_src = inputfile.read()
            inputfile.close()
            self.parse(adoc_src)

    def parse(self, data: bytes) -> None:
        """Process the given source string (binary)."""
        text = data.decode()
        lines = text.splitlines(keepends=True)

        # Parse document header if present
        lines = self._parse_header(lines)

        # Parse the rest of the document
        self._parse_content(lines)

        # Reconstruct the document
        self._reconstruct()

    def _parse_header(self, lines: list[str]) -> list[str]:
        """
        Parse document header if present.

        Returns remaining lines after header.
        """
        # Check for document header (first line starting with = )
        if not lines or not lines[0].startswith("= "):
            return lines

        # The header includes: title line, optional author/revision lines, and attributes
        # Keep going until we hit a blank line followed by section/paragraph content
        header_end = 0  # At minimum, include the title
        seen_blank = False
        i = 1

        while i < len(lines):
            line = lines[i]

            # Check for comment block delimiter ////
            # Comment blocks can appear in the header and should be included
            if self._is_comment_block_delimiter(line):
                header_end, i = self._skip_comment_block(lines, i, header_end)
                continue

            # Attributes start with : - always part of header
            if line.startswith(":"):
                header_end = i
                seen_blank = False  # Reset blank line tracking
            # Empty line - might separate header sections or end header
            elif not line.strip():
                if seen_blank:
                    # Two blank lines in a row typically ends header
                    break
                seen_blank = True
                # Check if next line starts content
                if i + 1 < len(lines):
                    next_line = lines[i + 1]
                    # If next line is a section (==) or regular paragraph, end here
                    if next_line.strip() and (
                        next_line.startswith("==")
                        or (
                            not next_line.startswith(":")
                            and not self._is_comment_block_delimiter(next_line)
                        )
                    ):
                        header_end = i
                        break
                header_end = i
            # Non-empty, non-attribute line after title
            # Could be author, revision info, or start of content
            else:
                # If we've seen a blank line and this doesn't start with :
                # it's likely the start of content
                if seen_blank:
                    break
                # Check if this looks like author/revision line (typically line 1-3 after title)
                # Author lines don't start with [ or = and are early in the document
                if i <= 3 and not line.startswith("[") and not line.startswith("="):
                    header_end = i
                else:
                    # This is content, not header
                    break
            i += 1

        if header_end >= 0:
            header_content = "".join(lines[: header_end + 1])
            header = AsciiDocHeaderUnit(header_content)
            self.addunit(header)
            self._elements.append(
                {"type": "header", "content": header_content, "unit": header}
            )
            return lines[header_end + 1 :]

        return lines

    def _is_comment_block_delimiter(self, line: str) -> bool:
        """Check if a line is a comment block delimiter (////)."""
        return line.strip() == "////"

    def _skip_comment_block(
        self, lines: list[str], i: int, header_end: int
    ) -> tuple[int, int]:
        """Skip a comment block and return updated header_end and line index."""
        # Include the comment block in the header
        header_end = i
        i += 1
        # Skip until closing delimiter
        while i < len(lines) and not self._is_comment_block_delimiter(lines[i]):
            header_end = i
            i += 1
        if i < len(lines):
            header_end = i  # Include closing delimiter
            i += 1
        return header_end, i

    def _parse_content(self, lines: list[str]) -> None:
        """Parse AsciiDoc content and extract translation units."""
        i = 0
        while i < len(lines):
            line = lines[i]

            # Skip empty lines
            if not line.strip():
                self._elements.append({"type": "empty", "content": line})
                i += 1
                continue

            # Try each parsing method in order
            if self._try_parse_conditional(lines, i):
                i = self._get_next_index(lines, i, "conditional")
            elif (
                self._try_parse_directive(line, i)
                or self._try_parse_anchor(line, i)
                or self._try_parse_block_title(line, i)
                or self._try_parse_attribute(line, i)
                or self._try_parse_heading(line, i)
                or self._try_parse_unordered_list(line, i)
                or self._try_parse_ordered_list(line, i)
                or self._try_parse_description_list(line, i)
            ):
                i += 1
            elif self._try_parse_block_delimiter(lines, i):
                i = self._get_next_index(lines, i, "block")
            elif (
                self._try_parse_list_continuation(line, i)
                or self._try_parse_comment(line, i)
                or self._try_parse_admonition(line, i)
            ):
                i += 1
            elif self._try_parse_table(lines, i):
                i = self._get_next_index(lines, i, "table")
            else:
                # Parse as paragraph
                i = self._parse_paragraph(lines, i)

    def _try_parse_conditional(self, lines: list[str], i: int) -> bool:
        """Parse conditional directives (ifdef, ifndef, ifeval)."""
        line = lines[i]
        if not re.match(r"^(ifdef|ifndef|ifeval)::", line):
            return False

        # Store the opening directive
        block_lines = [line]
        i += 1
        # Collect everything until endif
        depth = 1
        while i < len(lines) and depth > 0:
            current_line = lines[i]
            block_lines.append(current_line)
            if re.match(r"^(ifdef|ifndef|ifeval)::", current_line):
                depth += 1
            elif re.match(r"^endif::", current_line):
                depth -= 1
                if depth == 0:
                    break
            i += 1
        self._elements.append(
            {
                "type": "conditional_block",
                "content": "".join(block_lines),
                "end_index": i + 1,
            }
        )
        return True

    def _try_parse_directive(self, line: str, i: int) -> bool:
        """Parse standalone endif directive."""
        if not re.match(r"^endif::", line):
            return False
        self._elements.append({"type": "directive", "content": line})
        return True

    def _try_parse_anchor(self, line: str, i: int) -> bool:
        """Parse anchor [[anchor-id]]."""
        # Match anchor syntax: [[id]] where id contains non-bracket characters
        # Pattern [^\]]+ efficiently matches typical anchor IDs (short identifiers)
        # without complex backtracking issues
        if not re.match(r"^\[\[[^\]]+\]\]\s*$", line):
            return False
        self._elements.append({"type": "anchor", "content": line})
        return True

    def _try_parse_block_title(self, line: str, i: int) -> bool:
        """Parse block title (starts with . followed by alphanumeric)."""
        if not re.match(r"^\.[A-Za-z0-9]", line):
            return False
        self._elements.append({"type": "block_title", "content": line})
        return True

    def _try_parse_attribute(self, line: str, i: int) -> bool:
        """Parse attribute lines (e.g., [NOTE], [source,java])."""
        if not (line.strip().startswith("[") and line.strip().endswith("]")):
            return False
        self._elements.append({"type": "attribute", "content": line})
        return True

    def _try_parse_heading(self, line: str, i: int) -> bool:
        """Parse section heading."""
        heading_match = re.match(r"^(={2,6})\s+(\S.*?)$", line)
        if not heading_match:
            return False

        level = len(heading_match.group(1))
        # Strip any trailing whitespace and closing = markers
        title = heading_match.group(2).rstrip().rstrip("=").strip()

        # Update heading hierarchy and get docpath
        docpath = self._enter_heading(level)

        unit = self.addsourceunit(title)
        unit.addlocation(f"{self.filename or ''}:{i + 1}")
        unit.setdocpath(docpath)
        unit.set_element_info(
            "heading",
            heading_match.group(1) + " ",
            "\n" if line.endswith("\n") else "",
        )

        self._elements.append(
            {
                "type": "heading",
                "level": level,
                "prefix": heading_match.group(1) + " ",
                "suffix": "\n" if line.endswith("\n") else "",
                "unit": unit,
                "line": i + 1,
            }
        )
        return True

    def _try_parse_unordered_list(self, line: str, i: int) -> bool:
        """Parse unordered list item."""
        list_match = re.match(r"^(\*+)\s+(\S.*?)$", line)
        if not list_match:
            return False

        level = len(list_match.group(1))
        content = list_match.group(2).strip()

        # Handle checklist syntax [*], [x], [ ]
        checklist_prefix = ""
        # Use .*? for non-greedy matching to avoid backtracking
        checklist_match = re.match(r"^(\[[*x ]\])\s+(\S.*?)$", content)
        if checklist_match:
            checklist_prefix = checklist_match.group(1) + " "
            content = checklist_match.group(2).strip()

        docpath = self._build_docpath("li")
        unit = self.addsourceunit(content)
        unit.addlocation(f"{self.filename or ''}:{i + 1}")
        unit.setdocpath(docpath)
        prefix = list_match.group(1) + " " + checklist_prefix
        unit.set_element_info(
            "list_item",
            prefix,
            "\n" if line.endswith("\n") else "",
        )

        self._elements.append(
            {
                "type": "list_item",
                "level": level,
                "prefix": prefix,
                "suffix": "\n" if line.endswith("\n") else "",
                "unit": unit,
                "line": i + 1,
            }
        )
        return True

    def _try_parse_ordered_list(self, line: str, i: int) -> bool:
        """Parse ordered list item."""
        ordered_list_match = re.match(r"^(\.+)\s+(\S.*?)$", line)
        if not ordered_list_match:
            return False

        level = len(ordered_list_match.group(1))
        content = ordered_list_match.group(2).strip()

        docpath = self._build_docpath("li")
        unit = self.addsourceunit(content)
        unit.addlocation(f"{self.filename or ''}:{i + 1}")
        unit.setdocpath(docpath)
        unit.set_element_info(
            "list_item",
            ordered_list_match.group(1) + " ",
            "\n" if line.endswith("\n") else "",
        )

        self._elements.append(
            {
                "type": "list_item",
                "level": level,
                "prefix": ordered_list_match.group(1) + " ",
                "suffix": "\n" if line.endswith("\n") else "",
                "unit": unit,
                "line": i + 1,
            }
        )
        return True

    def _try_parse_description_list(self, line: str, i: int) -> bool:
        """Parse description list (term:: definition)."""
        # Use [^:\n]* to match non-colon/non-newline characters for the term.
        # This prevents the term capture from consuming the :: delimiter when
        # the input contains multiple colons before the actual delimiter.
        desc_list_match = re.match(r"^(\S[^:\n]*)::\s+(\S.*?)$", line)
        if not desc_list_match:
            return False

        term = desc_list_match.group(1).strip()
        definition = desc_list_match.group(2).strip()

        # Create unit for the definition only (term is part of the markup)
        docpath = self._build_docpath("dl")
        unit = self.addsourceunit(definition)
        unit.addlocation(f"{self.filename or ''}:{i + 1}")
        unit.setdocpath(docpath)
        unit.set_element_info(
            "description_list",
            f"{term}:: ",
            "\n" if line.endswith("\n") else "",
        )

        self._elements.append(
            {
                "type": "description_list",
                "term": term,
                "prefix": f"{term}:: ",
                "suffix": "\n" if line.endswith("\n") else "",
                "unit": unit,
                "line": i + 1,
            }
        )
        return True

    def _try_parse_block_delimiter(self, lines: list[str], i: int) -> bool:
        """Parse code blocks, literal blocks, example blocks, etc."""
        line = lines[i]
        if not (
            line.strip()
            and len(set(line.strip())) == 1
            and line.strip()[0] in "-=.*_+/"
        ):
            return False

        # Check if it's a delimiter (4+ repeated characters)
        delimiter = line.strip()
        if len(delimiter) < 4:
            return False

        block_lines = [line]
        i += 1
        while i < len(lines):
            block_lines.append(lines[i])
            if lines[i].strip() == delimiter:
                i += 1
                break
            i += 1

        self._elements.append(
            {"type": "code_block", "content": "".join(block_lines), "end_index": i}
        )
        return True

    def _try_parse_list_continuation(self, line: str, i: int) -> bool:
        """Parse list continuation marker (standalone +)."""
        if line.strip() != "+":
            return False
        self._elements.append({"type": "list_continuation", "content": line})
        return True

    def _try_parse_comment(self, line: str, i: int) -> bool:
        """Parse comment line."""
        if not line.startswith("//"):
            return False
        self._elements.append({"type": "comment", "content": line})
        return True

    def _try_parse_admonition(self, line: str, i: int) -> bool:
        """Parse admonition (NOTE, TIP, IMPORTANT, WARNING, CAUTION)."""
        admonition_match = re.match(
            r"^(NOTE|TIP|IMPORTANT|WARNING|CAUTION):\s+(\S.*?)$", line
        )
        if not admonition_match:
            return False

        admon_type = admonition_match.group(1)
        content = admonition_match.group(2).strip()

        docpath = self._build_docpath("admonition")
        unit = self.addsourceunit(content)
        unit.addlocation(f"{self.filename or ''}:{i + 1}")
        unit.setdocpath(docpath)
        unit.set_element_info(
            "admonition",
            f"{admon_type}: ",
            "\n" if line.endswith("\n") else "",
        )

        self._elements.append(
            {
                "type": "admonition",
                "admon_type": admon_type,
                "prefix": f"{admon_type}: ",
                "suffix": "\n" if line.endswith("\n") else "",
                "unit": unit,
                "line": i + 1,
            }
        )
        return True

    def _try_parse_table(self, lines: list[str], i: int) -> bool:
        """Parse table."""
        line = lines[i]
        if not line.strip().startswith("|"):
            return False

        table_lines = []
        start_line = i
        # Collect all table lines
        while i < len(lines) and (
            lines[i].strip().startswith("|") or not lines[i].strip()
        ):
            table_lines.append(lines[i])
            i += 1
            if i < len(lines) and not lines[i].strip():
                # Check if next non-empty line is still part of table
                next_i = i
                while next_i < len(lines) and not lines[next_i].strip():
                    next_i += 1
                if next_i >= len(lines) or not lines[next_i].strip().startswith("|"):
                    break

        # Parse table cells for translation
        # Note: This is a simple pipe-based split that doesn't handle:
        # - Escaped pipes (e.g., \|)
        # - CSV/PSV/DSV table formats (comma/colon/delimiter-separated values)
        # - Cell spanning (colspan/rowspan) with +/. notation
        # - Formatted table cells (a, e, h, l, m, s, v cell specifiers)
        # - Complex table features (cols/rows attributes, cell alignment)
        # For more sophisticated table parsing, consider using an AsciiDoc parser library.

        # Build table docpath
        table_docpath = self._build_docpath("table")
        row_index = 0

        for table_line in table_lines:
            if table_line.strip() and "|" in table_line:
                row_index += 1
                # Extract cells (simple approach)
                cells = [cell.strip() for cell in table_line.split("|") if cell.strip()]
                for col_index, cell in enumerate(cells, 1):
                    # Skip cell separator markers and empty cells
                    if cell and not cell.startswith("="):
                        cell_docpath = f"{table_docpath}/r[{row_index}]/c[{col_index}]"
                        unit = self.addsourceunit(cell)
                        unit.addlocation(f"{self.filename or ''}:{start_line + 1}")
                        unit.setdocpath(cell_docpath)
                        unit.set_element_info("table_cell", "", "")

        self._elements.append(
            {
                "type": "table",
                "content": "".join(table_lines),
                "line": start_line + 1,
                "end_index": i,
            }
        )
        return True

    def _parse_paragraph(self, lines: list[str], i: int) -> int:
        """Parse paragraph - collect consecutive non-empty lines."""
        para_lines = []
        start_line = i
        while i < len(lines) and lines[i].strip():
            # Check if this is a special line that breaks paragraphs
            line_stripped = lines[i].strip()
            # Check for block delimiters (4+ repeated characters)
            # Include '/' for comment blocks (////) to maintain consistency with
            # _try_parse_block_delimiter. Comment delimiters should properly break
            # paragraph parsing just like other block delimiters.
            is_delimiter = (
                len(line_stripped) >= 4
                and len(set(line_stripped)) == 1
                and line_stripped[0] in "-=.*_+/"
            )
            if (
                re.match(r"^(={2,6}|\*+|\.+)\s+", lines[i])
                or is_delimiter
                or lines[i].startswith("//")
            ):
                break
            para_lines.append(lines[i])
            i += 1

        if para_lines:
            # Join paragraph lines and normalize whitespace
            para_text = " ".join(line.strip() for line in para_lines)

            if para_text:
                docpath = self._build_docpath("p")
                unit = self.addsourceunit(para_text)
                unit.addlocation(f"{self.filename or ''}:{start_line + 1}")
                unit.setdocpath(docpath)
                unit.set_element_info("paragraph", "", "\n")

                self._elements.append(
                    {
                        "type": "paragraph",
                        "unit": unit,
                        "line": start_line + 1,
                        "original_lines": para_lines,
                    }
                )
        return i

    def _get_next_index(
        self, lines: list[str], current_i: int, element_type: str
    ) -> int:
        """Get the next index after parsing a multi-line element."""
        # Find the last element added and get its end_index
        if self._elements and "end_index" in self._elements[-1]:
            return self._elements[-1]["end_index"]
        return current_i + 1

    def _reconstruct(self) -> None:
        """Reconstruct the AsciiDoc document with translations."""
        result = []

        for element in self._elements:
            elem_type = element["type"]

            if elem_type == "header":
                if element.get("unit") and element["unit"].isheader():
                    result.append(element["content"])
            elif elem_type in {
                "empty",
                "code_block",
                "comment",
                "attribute",
                "list_continuation",
                "directive",
                "anchor",
                "block_title",
                "conditional_block",
            }:
                result.append(element["content"])
            elif elem_type in {
                "heading",
                "list_item",
                "admonition",
                "description_list",
            }:
                unit = element.get("unit")
                if unit:
                    translated = self.callback(unit.source)
                    result.append(f"{element['prefix']}{translated}{element['suffix']}")
            elif elem_type == "table":
                # For tables, we need to reconstruct with translated cells
                # For now, preserve the original table structure
                # (proper table translation would require more complex parsing)
                result.append(element["content"])
            elif elem_type == "paragraph":
                unit = element.get("unit")
                if unit:
                    translated = self.callback(unit.source)
                    # Try to preserve line structure somewhat
                    result.append(f"{translated}\n")

        self.filesrc = "".join(result)

    @staticmethod
    def _dummy_callback(text: str) -> str:
        """Default callback that returns text unchanged."""
        return text

    def _build_docpath(self, element_type: str) -> str:
        """
        Build the current docpath string for a block element.

        Docpath format follows the heading hierarchy with element counters:
        h2[1]/h3[1]/p[2] - second paragraph under first h3 under first h2
        h2[1]/table[1]/r[1]/c[2] - second column of first row of first table under first h2
        """
        parts = [f"h{level}[{idx}]" for level, idx in self._heading_stack]
        counts = self._section_counts[-1]
        counts[element_type] = counts.get(element_type, 0) + 1
        parts.append(f"{element_type}[{counts[element_type]}]")
        return "/".join(parts)

    def _enter_heading(self, level: int) -> str:
        """
        Update heading stack when entering a heading and return its docpath.

        Pops any headings of equal or greater level, counts this heading at the
        current section level, and pushes a new section level.
        """
        # Pop headings of equal or greater level
        while self._heading_stack and self._heading_stack[-1][0] >= level:
            self._heading_stack.pop()
            self._section_counts.pop()
        # Count this heading at the current section level
        counts = self._section_counts[-1]
        heading_key = f"h{level}"
        counts[heading_key] = counts.get(heading_key, 0) + 1
        idx = counts[heading_key]
        # Push new section level
        self._heading_stack.append((level, idx))
        self._section_counts.append({})
        # Build and return path
        return "/".join(f"h{lvl}[{i}]" for lvl, i in self._heading_stack)



#
# Copyright 2006-2007 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert Comma-Separated Value (.csv) files to a TermBase eXchange (.tbx)
glossary file.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/csv2tbx.html
for examples and usage instructions
"""

from translate.convert import convert
from translate.storage import csvl10n, tbx


class csv2tbx:
    """
    a class that takes translations from a .csv file and puts them in a .tbx
    file.
    """

    def __init__(self, charset=None) -> None:
        """Construct the converter..."""
        self.charset = charset

    def convertfile(self, csvfile):
        """
        Converts a csvfile to a tbxfile, and returns it. uses templatepo if
        given at construction.
        """
        mightbeheader = True
        self.tbxfile = tbx.tbxfile()
        for csvunit in csvfile.units:
            if mightbeheader:
                # ignore typical header strings...
                mightbeheader = False
                if csvunit.match_header():
                    continue
                if (
                    len(csvunit.location.strip()) == 0
                    and csvunit.source.find("Content-Type:") != -1
                ):
                    continue
            term = tbx.tbxunit.buildfromunit(csvunit)
            # TODO: we might want to get the location or other information
            # from CSV
            self.tbxfile.addunit(term)
        return self.tbxfile


def convertcsv(
    inputfile, outputfile, templatefile, charset=None, columnorder=None
) -> int:
    """
    Reads in inputfile using csvl10n, converts using csv2tbx, writes to
    outputfile.
    """
    inputstore = csvl10n.csvfile(inputfile, fieldnames=columnorder)
    convertor = csv2tbx(charset=charset)
    outputstore = convertor.convertfile(inputstore)
    if len(outputstore.units) == 0:
        return 0
    outputstore.serialize(outputfile)
    return 1


def main() -> None:
    formats = {
        ("csv", "tbx"): ("tbx", convertcsv),
        ("csv", None): ("tbx", convertcsv),
    }
    parser = convert.ConvertOptionParser(
        formats, usetemplates=False, description=__doc__
    )
    parser.add_option(
        "",
        "--charset",
        dest="charset",
        default=None,
        help="set charset to decode from csv files",
        metavar="CHARSET",
    )
    parser.add_option(
        "",
        "--columnorder",
        dest="columnorder",
        default=None,
        help="specify the order and position of columns (comment,source,target)",
    )
    parser.passthrough.append("charset")
    parser.passthrough.append("columnorder")
    parser.run()


if __name__ == "__main__":
    main()



#
# Copyright 2024 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Conservative MDX storage support built on the Markdown storage."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from translate.storage.markdown import MarkdownFile, MarkdownUnit

if TYPE_CHECKING:
    from collections.abc import Iterator

_COMPONENT_NAME = r"[A-Z][A-Za-z0-9_$]*(?:\.[A-Za-z_$][A-Za-z0-9_$]*)*"
_ATTRIBUTE_NAME = r"[A-Za-z_:][A-Za-z0-9_.:-]*"
_QUOTED_VALUE = r'(?:"[^"\r\n]*"|\'[^\'\r\n]*\')'
_ATTRIBUTE = rf"\s+{_ATTRIBUTE_NAME}(?:\s*=\s*{_QUOTED_VALUE})?"

_COMPONENT_START = re.compile(rf"^<(?P<name>{_COMPONENT_NAME})(?=[\s/>]|$)")
_COMPONENT_CLOSE = re.compile(rf"^</(?P<name>{_COMPONENT_NAME})\s*>[ \t]*$")
_FRAGMENT_START = re.compile(r"^(?:<>|</>)")
_SIMPLE_TAG = re.compile(
    rf"^<(?P<name>{_COMPONENT_NAME})(?P<attrs>(?:{_ATTRIBUTE})*)"
    rf"\s*(?P<slash>/?)>[ \t]*$"
)
_ATTRIBUTE_TOKEN = re.compile(
    rf"\s+(?P<name>{_ATTRIBUTE_NAME})"
    rf"(?:\s*=\s*(?:\"(?P<double>[^\"\r\n]*)\"|'(?P<single>[^'\r\n]*)'))?"
)
_ESM_START = re.compile(
    r"""^(?:
        import \s+ (?:
            \{|\*|type\s|\"|\'|`
            |[A-Za-z_$][A-Za-z0-9_$]*\s*,\s*(?:\*|\{)
            |[A-Za-z_$][A-Za-z0-9_$]*\s+from\b
        )
        |
        export \s+ (?:default|const|function|class|type|let|var|\{|\*)
    )""",
    re.VERBOSE,
)
_FENCE = re.compile(r"^ {0,3}(?P<delimiter>`{3,}|~{3,})")
_LOWER_HTML_START = re.compile(r"^ {0,3}<(?P<tag>[a-z][A-Za-z0-9-]*)(?:[\s>/]|$)")


class MDXUnit(MarkdownUnit):
    """A unit of translatable/localisable MDX content."""


class MDXFile(MarkdownFile):
    """Parse a deliberately small, structurally isolated MDX subset."""

    UnitClass = MDXUnit

    def __init__(
        self,
        inputfile=None,
        callback=None,
        max_line_length=None,
        extract_code_blocks=True,
        extract_frontmatter=True,
        frontmatter_translate_values=False,
        no_placeholders=False,
    ) -> None:
        """Construct an MDX store with the same options as MarkdownFile."""
        self._mdx_blocks: dict[str, tuple[str, int]] = {}
        self._mdx_block_counter = 0
        self._mdx_block_placeholder_prefix = "MDX_BLOCK"
        self._jsx_attr_counter = 0
        self._jsx_block_counter = 0
        super().__init__(
            inputfile=inputfile,
            callback=callback,
            max_line_length=max_line_length,
            extract_code_blocks=extract_code_blocks,
            extract_frontmatter=extract_frontmatter,
            frontmatter_translate_values=frontmatter_translate_values,
            no_placeholders=no_placeholders,
        )

    def parse(self, data) -> None:
        """Protect MDX syntax, parse Markdown, and restore protected content."""
        self._mdx_blocks = {}
        self._mdx_block_counter = 0
        self._jsx_attr_counter = 0
        self._jsx_block_counter = 0

        text = data.decode() if isinstance(data, bytes) else data
        self._mdx_block_placeholder_prefix = self._choose_placeholder_prefix(text)
        frontmatter_offset = self._frontmatter_body_offset(text.split("\n"))
        super().parse(self._protect_mdx(text).encode())
        self.units.sort(
            key=lambda unit: self._unit_document_line(unit, frontmatter_offset)
        )
        self.filesrc = self._restore_mdx_blocks(self.filesrc)

    def _protect_mdx(self, text: str) -> str:
        """Replace recognized MDX blocks with line-count-preserving comments."""
        lines = text.split("\n")
        frontmatter_lines = self._frontmatter_line_count(lines)
        result = lines[:frontmatter_lines]
        i = frontmatter_lines
        fence: tuple[str, int] | None = None

        while i < len(lines):
            line = lines[i]

            if fence is not None:
                result.append(line)
                if self._is_fence_close(line, fence):
                    fence = None
                i += 1
                continue

            opening_fence = self._fence_open(line)
            if opening_fence is not None:
                fence = opening_fence
                result.append(line)
                i += 1
                continue

            if self._is_indented_code(line):
                result.append(line)
                i += 1
                continue

            if not line.strip():
                result.append(line)
                i += 1
                continue

            if line.strip() == "<!-- translate:off -->":
                end = self._translate_off_end(lines, i)
                result.append(self._opaque_lines(lines, i, end))
                i = end
                continue

            raw_end = self._raw_html_end(lines, i)
            if raw_end is not None:
                result.append(self._opaque_lines(lines, i, raw_end))
                i = raw_end
                continue

            if self._is_esm_start(line):
                end = self._blank_delimited_end(lines, i)
                result.append(self._opaque_lines(lines, i, end))
                i = end
                continue

            component = _COMPONENT_START.match(line)
            if component is not None:
                end, rendered = self._component_block(lines, i, component.group("name"))
                result.append(
                    self._make_placeholder(rendered, source_line_count=end - i)
                )
                i = end
                continue

            if _COMPONENT_CLOSE.fullmatch(
                self._logical_line(line)
            ) or _FRAGMENT_START.match(self._logical_line(line)):
                end = self._extend_to_blank(lines, i + 1)
                result.append(self._opaque_lines(lines, i, end))
                i = end
                continue

            end = self._blank_delimited_end(lines, i)
            if self._chunk_requires_opaque(lines[i:end]):
                result.append(self._opaque_lines(lines, i, end))
            else:
                result.extend(lines[i:end])
            i = end

        return "\n".join(result)

    def _component_block(
        self, lines: list[str], start: int, name: str
    ) -> tuple[int, str]:
        """Return the end and rendering of one column-zero component block."""
        opening = lines[start]
        simple = _SIMPLE_TAG.fullmatch(self._logical_line(opening))
        if simple is not None and simple.group("slash"):
            end = start + 1
            if self._is_block_end(lines, end):
                return end, self._translate_simple_tag(opening, start + 1)
            end = self._extend_to_blank(lines, end)
            return end, "\n".join(lines[start:end])

        first_close = self._first_standalone_close(lines, start + 1, name)
        if (
            simple is not None
            and first_close is not None
            and self._is_block_end(lines, first_close + 1)
        ):
            child_lines = lines[start + 1 : first_close]
            if self._simple_markdown_children(child_lines):
                return first_close + 1, self._render_simple_component(
                    opening,
                    child_lines,
                    lines[first_close],
                    start,
                )

        end = self._opaque_component_end(lines, start, name)
        end = self._extend_to_blank(lines, end)
        return end, "\n".join(lines[start:end])

    def _render_simple_component(
        self,
        opening: str,
        child_lines: list[str],
        closing: str,
        opening_index: int,
    ) -> str:
        """Translate one simple component as an isolated Markdown subdocument."""
        translated_opening = self._translate_simple_tag(opening, opening_index + 1)
        self._jsx_block_counter += 1
        if not child_lines:
            return f"{translated_opening}\n{closing}"

        indent = self._common_child_indent(child_lines)
        dedented_lines = [line[indent:] for line in child_lines]
        child_store = MarkdownFile(
            callback=self.callback,
            max_line_length=self.max_line_length,
            extract_code_blocks=self.extract_code_blocks,
            extract_frontmatter=False,
            no_placeholders=self.no_placeholders,
        )
        child_store.parse("\n".join(dedented_lines).encode())

        self._copy_child_units(child_store, opening_index + 2)
        rendered_children = self._reindent(child_store.filesrc, " " * indent)
        if rendered_children and not rendered_children.endswith("\n"):
            rendered_children += "\n"
        return f"{translated_opening}\n{rendered_children}{closing}"

    def _copy_child_units(self, child_store: MarkdownFile, first_line: int) -> None:
        """Copy isolated Markdown units into this store with absolute locations."""
        path_prefix = f"jsx-block[{self._jsx_block_counter}]"
        for child_unit in child_store.units:
            unit = self.addsourceunit(child_unit.source)
            relative_line = self._unit_source_line(child_unit) or 1
            unit.addlocation(f"{self.filename or ''}:{first_line + relative_line - 1}")
            child_path = child_unit.getdocpath()
            unit.setdocpath(
                f"{path_prefix}/{child_path}" if child_path else path_prefix
            )

    def _translate_simple_tag(self, tag: str, line_number: int) -> str:
        """Translate quoted values in a validated simple component tag."""
        logical_tag = self._logical_line(tag)
        match = _SIMPLE_TAG.fullmatch(logical_tag)
        if match is None:
            return tag

        result: list[str] = []
        pos = 0
        attrs_start, attrs_end = match.span("attrs")
        for attribute in _ATTRIBUTE_TOKEN.finditer(logical_tag, attrs_start, attrs_end):
            value_group = (
                "double" if attribute.group("double") is not None else "single"
            )
            if attribute.group(value_group) is None:
                continue
            value_start, value_end = attribute.span(value_group)
            quote = '"' if value_group == "double" else "'"
            result.append(tag[pos:value_start])
            translated = self._translate_attribute(
                tag[value_start:value_end],
                match.group("name"),
                attribute.group("name"),
                line_number,
            )
            result.append(
                self._escape_attribute(
                    translated,
                    quote,
                    changed=translated != tag[value_start:value_end],
                )
            )
            pos = value_end
        result.append(tag[pos:])
        return "".join(result)

    def _translate_attribute(
        self, value: str, component: str, attribute: str, line_number: int
    ) -> str:
        """Emit one quoted component attribute and return its translation."""
        self._jsx_attr_counter += 1
        if not value.strip():
            return value
        unit = self.addsourceunit(value)
        unit.addlocation(f"{self.filename or ''}:{line_number}")
        unit.setdocpath(f"jsx-attr[{self._jsx_attr_counter}]/{component}.@{attribute}")
        return self.callback(value)

    @staticmethod
    def _escape_attribute(value: str, quote: str, *, changed: bool) -> str:
        """Encode the active delimiter as a JSX-compatible character entity."""
        entity = "&quot;" if quote == '"' else "&apos;"
        value = value.replace(quote, entity)
        if changed:
            value = value.replace("<", "&lt;")
            trailing_backslashes = len(value) - len(value.rstrip("\\"))
            if trailing_backslashes % 2:
                value += "\\"
        return value

    @classmethod
    def _simple_markdown_children(cls, lines: list[str]) -> bool:
        """Accept child Markdown only when it contains no MDX structure."""
        nonblank = [line for line in lines if line.strip()]
        if not nonblank:
            return True
        if any("\t" in line[: len(line) - len(line.lstrip())] for line in nonblank):
            return False
        if cls._common_child_indent(lines) >= 4:
            return False
        if nonblank[0].lstrip().startswith("---"):
            return False
        if any(
            _LOWER_HTML_START.match(line) is not None
            or cls._raw_html_end(lines, index) is not None
            for index, line in enumerate(lines)
        ):
            return False

        for _, line in cls._structural_lines(lines):
            if (
                re.search(
                    r"</?[A-Z][A-Za-z0-9_$]*(?:\.[A-Za-z_$][A-Za-z0-9_$]*)*",
                    line,
                )
                or "<>" in line
                or "</>" in line
                or "{" in line
                or "}" in line
                or cls._is_esm_start(line.lstrip())
            ):
                return False
        return True

    @staticmethod
    def _common_child_indent(lines: list[str]) -> int:
        """Return the common leading-space indentation of nonblank children."""
        indents = [len(line) - len(line.lstrip(" ")) for line in lines if line.strip()]
        return min(indents, default=0)

    @staticmethod
    def _reindent(text: str, prefix: str) -> str:
        """Restore a component child's structural indentation after rendering."""
        if not prefix:
            return text
        return "".join(
            f"{prefix}{line}" if line.strip() else line
            for line in text.splitlines(keepends=True)
        )

    @classmethod
    def _first_standalone_close(
        cls, lines: list[str], start: int, name: str
    ) -> int | None:
        """Find the first exact, column-zero closing line for a component."""
        for index, line in cls._structural_lines(lines, start):
            closing = _COMPONENT_CLOSE.fullmatch(line)
            if closing is not None and closing.group("name") == name:
                return index
        return None

    @classmethod
    def _opaque_component_end(cls, lines: list[str], start: int, name: str) -> int:
        """Conservatively collect unsupported component syntax as one block."""
        depth = 1
        for index, line in cls._structural_lines(lines, start + 1):
            if cls._is_opaque_same_name_open(line, name):
                depth += 1
                continue
            closing = _COMPONENT_CLOSE.fullmatch(line)
            if closing is not None and closing.group("name") == name:
                depth -= 1
                if depth == 0:
                    return index + 1
        return cls._blank_delimited_end(lines, start)

    @classmethod
    def _structural_lines(
        cls, lines: list[str], start: int = 0
    ) -> Iterator[tuple[int, str]]:
        """Yield logical lines outside fenced code and raw HTML blocks."""
        fence: tuple[str, int] | None = None
        index = start
        while index < len(lines):
            line = lines[index]
            if fence is not None:
                if cls._is_fence_close(line, fence):
                    fence = None
                index += 1
                continue
            fence = cls._fence_open(line)
            if fence is not None:
                index += 1
                continue
            raw_html_end = cls._raw_html_end(lines, index)
            if raw_html_end is not None:
                index = raw_html_end
                continue
            yield index, cls._logical_line(line)
            index += 1

    @staticmethod
    def _is_opaque_same_name_open(line: str, name: str) -> bool:
        """Recognize a complete same-name opening line without parsing its attrs."""
        opening = _COMPONENT_START.match(line)
        if opening is None or opening.group("name") != name:
            return False
        stripped = line.rstrip()
        if not stripped.endswith(">") or stripped.endswith("/>"):
            return False
        closing = re.compile(rf"</\s*{re.escape(name)}\s*>")
        return closing.search(line, opening.end()) is None

    @staticmethod
    def _logical_line(line: str) -> str:
        """Remove the CR from a CRLF line for syntax recognition."""
        return line.removesuffix("\r")

    def _opaque_lines(self, lines: list[str], start: int, end: int) -> str:
        """Create a placeholder for a source line range."""
        return self._make_placeholder(
            "\n".join(lines[start:end]), source_line_count=end - start
        )

    def _make_placeholder(
        self, content: str, source_line_count: int | None = None
    ) -> str:
        """Create a collision-resistant HTML comment placeholder."""
        self._mdx_block_counter += 1
        key = f"<!-- {self._mdx_block_placeholder_prefix}_{self._mdx_block_counter} -->"
        padding_lines = (
            content.count("\n")
            if source_line_count is None
            else max(source_line_count - 1, 0)
        )
        self._mdx_blocks[key] = (content, padding_lines)
        return key + ("\n" * padding_lines)

    def _restore_mdx_blocks(self, text: str) -> str:
        """Restore all opaque and rendered MDX blocks."""
        for key, (content, padding_lines) in self._mdx_blocks.items():
            text = text.replace(key + ("\n" * padding_lines), content)
            text = text.replace(key, content)
        return text

    @staticmethod
    def _choose_placeholder_prefix(text: str) -> str:
        """Choose a placeholder prefix not present in user content."""
        prefix = "MDX_BLOCK"
        counter = 1
        while f"<!-- {prefix}_" in text:
            prefix = f"MDX_BLOCK_{counter}"
            counter += 1
        return prefix

    @staticmethod
    def _frontmatter_line_count(lines: list[str]) -> int:
        """Return the number of leading YAML front matter lines."""
        if not lines or not lines[0].startswith("---"):
            return 0
        for line_number, line in enumerate(lines[1:], start=1):
            if line.startswith(("---", "...")):
                return line_number + 1
        return 0

    @classmethod
    def _frontmatter_body_offset(cls, lines: list[str]) -> int:
        """Return the line offset used by Markdown body unit locations."""
        offset = cls._frontmatter_line_count(lines)
        if offset and offset < len(lines) and not lines[offset].strip():
            offset += 1
        return offset

    @staticmethod
    def _blank_delimited_end(lines: list[str], start: int) -> int:
        """Return the first blank line after a conservative opaque block."""
        end = start + 1
        while end < len(lines) and lines[end].strip():
            end += 1
        return end

    @classmethod
    def _extend_to_blank(cls, lines: list[str], end: int) -> int:
        """Extend a non-isolated construct through its Markdown block."""
        if cls._is_block_end(lines, end):
            return end
        return cls._blank_delimited_end(lines, end)

    @staticmethod
    def _is_block_end(lines: list[str], end: int) -> bool:
        """Return whether a construct ends at EOF or before a blank line."""
        return end >= len(lines) or not lines[end].strip()

    @classmethod
    def _chunk_requires_opaque(cls, lines: list[str]) -> bool:
        """Reject an ordinary Markdown block containing structural MDX."""
        for line in lines:
            logical_line = cls._logical_line(line)
            if (
                _COMPONENT_START.match(logical_line)
                or _COMPONENT_CLOSE.fullmatch(logical_line)
                or _FRAGMENT_START.match(logical_line)
                or cls._is_esm_start(line)
            ):
                return True
            if (
                not cls._is_indented_code(line)
                and not re.match(r"(?:>|[-+*]\s|\d+[.)]\s)", line.lstrip())
                and ("{" in line or "}" in line)
            ):
                return True
        return False

    @staticmethod
    def _is_indented_code(line: str) -> bool:
        """Return whether a line has top-level Markdown code indentation."""
        column = 0
        for character in line:
            if character == " ":
                column += 1
            elif character == "\t":
                column += 4 - (column % 4)
            else:
                break
        return column >= 4

    @staticmethod
    def _translate_off_end(lines: list[str], start: int) -> int:
        """Collect an exact translation-control region through translate:on."""
        for index in range(start + 1, len(lines)):
            if lines[index].strip() == "<!-- translate:on -->":
                return index + 1
        return len(lines)

    @staticmethod
    def _is_esm_start(line: str) -> bool:
        """Recognize a top-level MDX ESM statement without parsing JavaScript."""
        return line == line.lstrip() and _ESM_START.match(line) is not None

    @staticmethod
    def _fence_open(line: str) -> tuple[str, int] | None:
        """Return the character and length of a top-level Markdown fence."""
        match = _FENCE.match(line)
        if match is None:
            return None
        delimiter = match.group("delimiter")
        return delimiter[0], len(delimiter)

    @staticmethod
    def _is_fence_close(line: str, fence: tuple[str, int]) -> bool:
        """Return whether a line closes the active Markdown fence."""
        character, length = fence
        match = re.fullmatch(r" {0,3}(`{3,}|~{3,})\s*", line)
        return bool(
            match and match.group(1)[0] == character and len(match.group(1)) >= length
        )

    @classmethod
    def _raw_html_end(cls, lines: list[str], start: int) -> int | None:
        """Return the end of raw HTML that must remain opaque to JSX scanning."""
        line = lines[start]
        stripped = line.lstrip(" ")
        if len(line) - len(stripped) > 3:
            return None

        if stripped.startswith("<!--"):
            return cls._marker_end(lines, start, "-->")
        if stripped.startswith("<?"):
            return cls._marker_end(lines, start, "?>")
        if stripped.startswith("<![CDATA["):
            return cls._marker_end(lines, start, "]]>")
        if re.match(r"<![A-Za-z]", stripped):
            return cls._marker_end(lines, start, ">")

        match = _LOWER_HTML_START.match(line)
        if match is None:
            return None
        tag = match.group("tag")
        if re.search(rf"</\s*{re.escape(tag)}\s*>", line, re.IGNORECASE):
            return None
        if line.rstrip().endswith("/>"):
            return None
        if tag.lower() in {"pre", "script", "style", "textarea"}:
            return cls._marker_end(lines, start, f"</{tag}>", ignore_case=True)
        return cls._blank_delimited_end(lines, start)

    @staticmethod
    def _marker_end(
        lines: list[str],
        start: int,
        marker: str,
        *,
        ignore_case: bool = False,
    ) -> int:
        """Collect through a literal raw-HTML terminator or end of file."""
        needle = marker.lower() if ignore_case else marker
        for index in range(start, len(lines)):
            haystack = lines[index].lower() if ignore_case else lines[index]
            if needle in haystack:
                return index + 1
        return len(lines)

    @staticmethod
    def _unit_source_line(unit: MarkdownUnit) -> int:
        """Return the numeric source line used for stable unit ordering."""
        for location in unit.getlocations():
            match = re.search(r":(\d+)$", location)
            if match:
                return int(match.group(1))
        return 0

    @classmethod
    def _unit_document_line(cls, unit: MarkdownUnit, frontmatter_offset: int) -> int:
        """Return an absolute source line for document-order sorting."""
        line = cls._unit_source_line(unit)
        docpath = unit.getdocpath() or ""
        if docpath.startswith("frontmatter"):
            return 0
        if docpath.startswith("jsx-"):
            return line
        return line + frontmatter_offset



#
# Copyright 2008, 2011 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

# Original Author: Dan Schafer <dschafer@mozilla.com>
# Date: 10 Jun 2008

"""Convert Mozilla .lang files to Gettext PO localization files."""

from translate.convert import convert
from translate.storage import mozilla_lang as lang
from translate.storage import po


class lang2po:
    """Convert one Mozilla .lang file to a single PO file."""

    SourceStoreClass = lang.LangStore
    TargetStoreClass = po.pofile
    TargetUnitClass = po.pounit

    def __init__(
        self,
        input_file,
        output_file,
        template_file=None,
        blank_msgstr=False,
        duplicate_style="msgctxt",
        encoding="utf-8",
    ) -> None:
        """Initialize the converter."""
        self.blank_msgstr = blank_msgstr
        self.duplicate_style = duplicate_style

        self.extraction_msg = None
        self.output_file = output_file
        self.source_store = self.SourceStoreClass(input_file, encoding=encoding)
        self.target_store = self.TargetStoreClass()
        self.template_store = None

        if template_file is not None:
            self.template_store = self.SourceStoreClass(template_file)

    def convert_unit(self, unit):
        """Convert a source format unit to a target format unit."""
        target_unit = self.TargetUnitClass()
        target_unit.addlocations(unit.getlocations())
        target_unit.addnote(unit.getnotes(), "developer")
        target_unit.source = unit.source
        target_unit.target = unit.target
        return target_unit

    def convert_store(self) -> None:
        """Convert a single source format file to a target format file."""
        self.extraction_msg = f"extracted from {self.source_store.filename}"

        for source_unit in self.source_store.units:
            self.target_store.addunit(self.convert_unit(source_unit))

    def merge_stores(self):
        """Convert two source format files to a target format file."""
        raise NotImplementedError

    def run(self) -> int:
        """Run the converter."""
        if self.template_store is None:
            self.convert_store()
        else:
            self.merge_stores()

        if self.extraction_msg:
            self.target_store.header().addnote(self.extraction_msg, "developer")

        self.target_store.removeduplicates(self.duplicate_style)

        if self.target_store.isempty():
            return 0

        self.target_store.serialize(self.output_file)
        return 1


def run_converter(
    input_file,
    output_file,
    template_file=None,
    pot=False,
    duplicatestyle="msgctxt",
    encoding="utf-8",
):
    """Wrapper around converter."""
    return lang2po(
        input_file,
        output_file,
        template_file,
        blank_msgstr=pot,
        duplicate_style=duplicatestyle,
        encoding=encoding,
    ).run()


formats = {"lang": ("po", run_converter)}


def main(argv=None) -> None:
    parser = convert.ConvertOptionParser(formats, usepots=True, description=__doc__)
    parser.add_option(
        "",
        "--encoding",
        dest="encoding",
        default="utf-8",
        help="The encoding of the input file (default: UTF-8)",
    )
    parser.passthrough.append("encoding")
    parser.add_duplicates_option()
    parser.run(argv)


if __name__ == "__main__":
    main()
